"""Djaizz TensorFlow.Keras Deep Learning Model class."""


from pathlib import Path
from sys import version_info

import h5py
from tensorflow.python.keras.saving.save import \
    load_model   # pylint: disable=no-name-in-module

from ...base import _AIModelWithArtifactFilesABC

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


__all__: Sequence[str] = ('KerasModel',)


class KerasModel(_AIModelWithArtifactFilesABC):
    # pylint: disable=abstract-method
    """Djaizz TensorFlow.Keras Deep Learning Model class."""

    class Meta(_AIModelWithArtifactFilesABC.Meta):
        # pylint: disable=too-few-public-methods
        """Django Model Class Metadata."""

        abstract = True

    def load(self) -> None:
        """Load the Model's native object from it artifact file."""
        if not self.native_obj:
            artifact_local_path = str(Path(self.artifact_local_path)
                                      .expanduser())

            self.native_obj = \
                load_model(filepath=(h5py.File(name=artifact_local_path,
                                               mode='r',
                                               driver=None,
                                               libver=None,
                                               userblock_size=None,
                                               swmr=False,
                                               rdcc_nslots=None,
                                               rdcc_nbytes=None,
                                               rdcc_w0=None,
                                               track_order=None,
                                               fs_strategy=None,
                                               fs_persist=False,
                                               fs_threshold=1)
                                     if artifact_local_path.endswith('.h5')
                                     else artifact_local_path),
                           custom_objects=None,
                           compile=True,
                           options=None)



"""
WSGI config for Djaizz project.

It exposes the WSGI callable as a module-level variable named ``application``

For more information on this file, see
docs.djangoproject.com/en/dev/howto/deployment/wsgi
"""


# ref: django-configurations.readthedocs.io


import os

# from django.core.wsgi import get_wsgi_application
from configurations.wsgi import get_wsgi_application


os.environ.setdefault(key='DJANGO_SETTINGS_MODULE', value='settings')
os.environ.setdefault(key='DJANGO_CONFIGURATION', value='Default')


application = get_wsgi_application()



"""Djaizz Model Views."""


from inspect import isclass
from typing import Literal, Union

from django.http.request import HttpRequest
from django.http.response import Http404, HttpResponse, HttpResponseRedirect
from django.shortcuts import redirect, render

from django_plotly_dash.dash_wrapper import DjangoDash
from django_plotly_dash.models import DashApp, StatelessApp

from gradio.inputs import Dropdown
from gradio.interface import Interface

from djaizz.model.models import AIModel


def model_ui(request: HttpRequest,
             model_class_or_instance_name_or_uuid: str,
             ui_type: Literal['dash', 'gradio']) \
        -> Union[Http404, HttpResponse, HttpResponseRedirect]:
    # pylint: disable=unused-argument
    """Launch AI Model class/instance's Dash or Gradio UI."""
    if model_class_or_instance_name_or_uuid in \
            (model_subclasses_by_name := AIModel.subclasses_by_name):
        # pylint: disable=trailing-whitespace,unsubscriptable-object
        if not (model_names_or_uuids :=   # noqa: W291
                (model := model_subclasses_by_name[
                    model_class_or_instance_name_or_uuid]).names_or_uuids):
            raise Http404('*** MODEL CLASS ' +
                          model_class_or_instance_name_or_uuid +
                          ' HAS NO INSTANCES ***')

    else:
        try:
            model: AIModel = AIModel.get_by_name_or_uuid(
                name_or_uuid=model_class_or_instance_name_or_uuid)

        except AIModel.DoesNotExist as ai_model_does_not_exist:
            raise Http404('*** MODEL INSTANCE ' +
                          model_class_or_instance_name_or_uuid +
                          ' NOT FOUND ***') from ai_model_does_not_exist

        model_names_or_uuids: list[str] = model.names_or_uuids

    if ui_type == 'dash':
        if not isinstance(dash_ui := model.dash_ui, DjangoDash):
            raise Http404(f'*** {model} DOES NOT HAVE A DASH UI ***')

        # if `model` is an AIModel instance, then render the view for its class
        if isinstance(model, AIModel):
            model_class_or_instance_name_or_uuid = type(model).__name__

        ((django_dash_stateless_app :=
          StatelessApp.objects.get_or_create(
            app_name=model_class_or_instance_name_or_uuid)[0])
         ._stateless_dash_app_instance) = dash_ui

        return render(
            request=request, template_name='Djaizz-Model-Dash-App.html',
            context={'django_dash_app':
                     DashApp.objects.get_or_create(
                        stateless_app=django_dash_stateless_app,
                        instance_name=model_class_or_instance_name_or_uuid)[0]
                     },
            content_type=None, status=None, using=None)

    if ui_type == 'gradio':
        if not isinstance(gradio_interface := model.gradio_ui, Interface):
            raise Http404(f'*** {model} DOES NOT HAVE A GRADIO UI ***')

        assert isinstance(gradio_interface.predict, list), \
            TypeError(f'*** {gradio_interface.predict} NOT A LIST ***')
        func: callable = gradio_interface.predict.pop()
        assert not gradio_interface.predict, \
            ValueError(f'*** {gradio_interface.predict} NOT AN EMPTY LIST ***')
        gradio_interface.predict.append(
            lambda model_name_or_uuid, *args:
            func(model.get_by_name_or_uuid(name_or_uuid=model_name_or_uuid),
                 *args))

        gradio_interface.input_components.insert(
            0,
            Dropdown(choices=model_names_or_uuids,
                     type='value',
                     default=(model_names_or_uuids[0]
                              if isclass(model)
                              else model.name_or_uuid),
                     label='AI Model Name or UUID'))

        _gradio_app, _gradio_path_to_local_server, gradio_share_url = \
            gradio_interface.launch(
                inline=False,
                # (bool) - whether to display in the interface inline
                # on python notebooks.

                inbrowser=True,
                # (bool) - whether to automatically launch the interface
                # in a new tab on the default browser.

                share=True,
                # (bool) - whether to create a publicly shareable link
                # from your computer for the interface.

                debug=False,
                # (bool) - if True, and the interface was launched
                # from Google Colab, prints the errors in the cell output.

                auth=None,
                # (Callable, Union[tuple[str, str], list[tuple[str, str]]]) -
                # If provided, username and password
                # (or list of username-password tuples)
                # required to access interface.
                # Can also provide function that takes username and password
                # and returns True if valid login.

                auth_message=None,
                # (str) - If provided, HTML message provided on login page.

                private_endpoint=None,
                prevent_thread_lock=False)

        return redirect(to=gradio_share_url, permanent=False)

    raise Http404('*** ui_type MUST BE EITHER "dash" or "gradio" ***')



"""Djaizz package."""


from importlib.metadata import version
from sys import version_info

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


__all__: Sequence[str] = ('__version__',)


__version__: str = version(distribution_name='Djaizz')



"""Djaizz Data API."""


from sys import version_info

from djaizz.data.models import (
    DataSchema, DataSet,

    InDBJSONDataSet, JSONDataSet,
    NumPyArray, PandasDataFrame,

    AudioDataSet,
    CSVDataSet,
    HDFDataSet,
    ImageDataSet,
    ORCDataSet,
    ParquetDataSet,
    TextDataSet,
    TFRecordDataSet,
    VideoDataSet,

    LiveAPIDataSource,
)

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


__all__: Sequence[str] = (
    'DataSchema', 'DataSet',

    'InDBJSONDataSet', 'JSONDataSet',
    'NumPyArray', 'PandasDataFrame',

    'AudioDataSet',
    'CSVDataSet',
    'HDFDataSet',
    'ImageDataSet',
    'ORCDataSet',
    'ParquetDataSet',
    'TextDataSet',
    'TFRecordDataSet',
    'VideoDataSet',

    'LiveAPIDataSource',
)



"""Set up pre-trained Hugging Face models."""


from transformers.pipelines import pipeline

from djaizz.model.models.ml.hugging_face import (
    PreTrainedHuggingFaceAudioClassifier,
    PreTrainedHuggingFaceImageClassifier,
    PreTrainedHuggingFaceMaskFiller,
    PreTrainedHuggingFaceObjectDetector,
    PreTrainedHuggingFaceQuestionAnswerer,
    PreTrainedHuggingFaceSpeechRecognizer,
    PreTrainedHuggingFaceTableQuestionAnswerer,
    PreTrainedHuggingFaceTextClassifier,
    PreTrainedHuggingFaceTextGenerator,
    PreTrainedHuggingFaceText2TextGenerator,
    PreTrainedHuggingFaceTextSummarizer,
    PreTrainedHuggingFaceTokenClassifier,
    PreTrainedHuggingFaceTranslator,
    PreTrainedHuggingFaceZeroShotClassifier,
)
from djaizz.util import full_qual_name


def run():   # noqa: MC0001
    """Run this script."""
    try:
        print(
            PreTrainedHuggingFaceAudioClassifier.objects.update_or_create(
                name='PreTrained-HuggingFace-Audio-Classifier',
                defaults=dict(
                    loader_module_and_qualname=full_qual_name(pipeline),
                    artifact_global_url=None,
                    artifact_local_path=None,
                    params=dict(__init__=dict(task='audio-classification',
                                              model=None,
                                              config=None,
                                              tokenizer=None,
                                              feature_extractor=None,
                                              framework=None,
                                              revision=None,
                                              use_fast=True,
                                              use_auth_token=None,
                                              model_kwargs={}))))[0])
    except Exception as err:   # pylint: disable=broad-except
        print(f'{PreTrainedHuggingFaceAudioClassifier.__name__} model '
              f'not set up: {err}')

    try:
        print(
            PreTrainedHuggingFaceImageClassifier.objects.update_or_create(
                name='PreTrained-HuggingFace-Image-Classifier',
                defaults=dict(
                    loader_module_and_qualname=full_qual_name(pipeline),
                    artifact_global_url=None,
                    artifact_local_path=None,
                    params=dict(__init__=dict(task='image-classification',
                                              model=None,
                                              config=None,
                                              tokenizer=None,
                                              feature_extractor=None,
                                              framework=None,
                                              revision=None,
                                              use_fast=True,
                                              use_auth_token=None,
                                              model_kwargs={}))))[0])
    except Exception as err:   # pylint: disable=broad-except
        print(f'{PreTrainedHuggingFaceImageClassifier.__name__} model '
              f'not set up: {err}')

    try:
        print(
            PreTrainedHuggingFaceMaskFiller.objects.update_or_create(
                name='PreTrained-HuggingFace-Mask-Filler',
                defaults=dict(
                    loader_module_and_qualname=full_qual_name(pipeline),
                    artifact_global_url=None,
                    artifact_local_path=None,
                    params=dict(__init__=dict(task='fill-mask',
                                              model=None,
                                              config=None,
                                              tokenizer=None,
                                              feature_extractor=None,
                                              framework=None,
                                              revision=None,
                                              use_fast=True,
                                              use_auth_token=None,
                                              model_kwargs={}))))[0])
    except Exception as err:   # pylint: disable=broad-except
        print(f'{PreTrainedHuggingFaceMaskFiller.__name__} model '
              f'not set up: {err}')

    try:
        print(
            PreTrainedHuggingFaceObjectDetector.objects.update_or_create(
                name='PreTrained-HuggingFace-Object-Detector',
                defaults=dict(
                    loader_module_and_qualname=full_qual_name(pipeline),
                    artifact_global_url=None,
                    artifact_local_path=None,
                    params=dict(__init__=dict(task='object-detection',
                                              model=None,
                                              config=None,
                                              tokenizer=None,
                                              feature_extractor=None,
                                              framework=None,
                                              revision=None,
                                              use_fast=True,
                                              use_auth_token=None,
                                              model_kwargs={}))))[0])
    except Exception as err:   # pylint: disable=broad-except
        print(f'{PreTrainedHuggingFaceObjectDetector.__name__} model '
              f'not set up: {err}')

    try:
        print(
            PreTrainedHuggingFaceQuestionAnswerer.objects.update_or_create(
                name='PreTrained-HuggingFace-Question-Answerer',
                defaults=dict(
                    loader_module_and_qualname=full_qual_name(pipeline),
                    artifact_global_url=None,
                    artifact_local_path=None,
                    params=dict(__init__=dict(task='question-answering',
                                              model=None,
                                              config=None,
                                              tokenizer=None,
                                              feature_extractor=None,
                                              framework=None,
                                              revision=None,
                                              use_fast=True,
                                              use_auth_token=None,
                                              model_kwargs={}))))[0])
    except Exception as err:   # pylint: disable=broad-except
        print(f'{PreTrainedHuggingFaceQuestionAnswerer.__name__} model '
              f'not set up: {err}')

    try:
        print(
            PreTrainedHuggingFaceSpeechRecognizer.objects.update_or_create(
                name='PreTrained-HuggingFace-Speech-Recognizer',
                defaults=dict(
                    loader_module_and_qualname=full_qual_name(pipeline),
                    artifact_global_url=None,
                    artifact_local_path=None,
                    params=dict(__init__=dict(task=('automatic-'
                                                    'speech-recognition'),
                                              model=None,
                                              config=None,
                                              tokenizer=None,
                                              feature_extractor=None,
                                              framework=None,
                                              revision=None,
                                              use_fast=True,
                                              use_auth_token=None,
                                              model_kwargs={}))))[0])
    except Exception as err:   # pylint: disable=broad-except
        print(f'{PreTrainedHuggingFaceSpeechRecognizer.__name__} model '
              f'not set up: {err}')

    try:
        print(
            PreTrainedHuggingFaceTableQuestionAnswerer.objects.update_or_create(   # noqa: E501
                name='PreTrained-HuggingFace-Table-Question-Answerer',
                defaults=dict(
                    loader_module_and_qualname=full_qual_name(pipeline),
                    artifact_global_url=None,
                    artifact_local_path=None,
                    params=dict(__init__=dict(task='table-question-answering',
                                              model=None,
                                              config=None,
                                              tokenizer=None,
                                              feature_extractor=None,
                                              framework=None,
                                              revision=None,
                                              use_fast=True,
                                              use_auth_token=None,
                                              model_kwargs={}))))[0])
    except Exception as err:   # pylint: disable=broad-except
        print(f'{PreTrainedHuggingFaceTableQuestionAnswerer.__name__} model '
              f'not set up: {err}')

    try:
        print(
            PreTrainedHuggingFaceTextClassifier.objects.update_or_create(
                name='PreTrained-HuggingFace-Text-Classifier',
                defaults=dict(
                    loader_module_and_qualname=full_qual_name(pipeline),
                    artifact_global_url=None,
                    artifact_local_path=None,
                    params=dict(__init__=dict(task='text-classification',
                                              model=None,
                                              config=None,
                                              tokenizer=None,
                                              feature_extractor=None,
                                              framework=None,
                                              revision=None,
                                              use_fast=True,
                                              use_auth_token=None,
                                              model_kwargs={}))))[0])
        print(
            PreTrainedHuggingFaceTextClassifier.objects.update_or_create(
                name='PreTrained-HuggingFace-Sentiment-Analyzer',
                defaults=dict(
                    loader_module_and_qualname=full_qual_name(pipeline),
                    artifact_global_url=None,
                    artifact_local_path=None,
                    params=dict(__init__=dict(task='sentiment-analysis',
                                              model=None,
                                              config=None,
                                              tokenizer=None,
                                              feature_extractor=None,
                                              framework=None,
                                              revision=None,
                                              use_fast=True,
                                              use_auth_token=None,
                                              model_kwargs={}))))[0])
    except Exception as err:   # pylint: disable=broad-except
        print(f'{PreTrainedHuggingFaceTextClassifier.__name__} model '
              f'not set up: {err}')

    try:
        print(
            PreTrainedHuggingFaceTextGenerator.objects.update_or_create(
                name='PreTrained-HuggingFace-Text-Generator',
                defaults=dict(
                    loader_module_and_qualname=full_qual_name(pipeline),
                    artifact_global_url=None,
                    artifact_local_path=None,
                    params=dict(__init__=dict(task='text-generation',
                                              model=None,
                                              config=None,
                                              tokenizer=None,
                                              feature_extractor=None,
                                              framework=None,
                                              revision=None,
                                              use_fast=True,
                                              use_auth_token=None,
                                              model_kwargs={}))))[0])
    except Exception as err:   # pylint: disable=broad-except
        print(f'{PreTrainedHuggingFaceTextGenerator.__name__} model '
              f'not set up: {err}')

    try:
        print(
            PreTrainedHuggingFaceText2TextGenerator.objects.update_or_create(
                name='PreTrained-HuggingFace-Text2Text-Generator',
                defaults=dict(
                    loader_module_and_qualname=full_qual_name(pipeline),
                    artifact_global_url=None,
                    artifact_local_path=None,
                    params=dict(__init__=dict(task='text2text-generation',
                                              model=None,
                                              config=None,
                                              tokenizer=None,
                                              feature_extractor=None,
                                              framework=None,
                                              revision=None,
                                              use_fast=True,
                                              use_auth_token=None,
                                              model_kwargs={}))))[0])
    except Exception as err:   # pylint: disable=broad-except
        print(f'{PreTrainedHuggingFaceText2TextGenerator.__name__} model '
              f'not set up: {err}')

    try:
        print(
            PreTrainedHuggingFaceTextSummarizer.objects.update_or_create(
                name='PreTrained-HuggingFace-Text-Summarizer',
                defaults=dict(
                    loader_module_and_qualname=full_qual_name(pipeline),
                    artifact_global_url=None,
                    artifact_local_path=None,
                    params=dict(__init__=dict(task='summarization',
                                              model=None,
                                              config=None,
                                              tokenizer=None,
                                              feature_extractor=None,
                                              framework=None,
                                              revision=None,
                                              use_fast=True,
                                              use_auth_token=None,
                                              model_kwargs={}))))[0])
    except Exception as err:   # pylint: disable=broad-except
        print(f'{PreTrainedHuggingFaceTextSummarizer.__name__} model '
              f'not set up: {err}')

    try:
        print(
            PreTrainedHuggingFaceTokenClassifier.objects.update_or_create(
                name='PreTrained-HuggingFace-Token-Classifier',
                defaults=dict(
                    loader_module_and_qualname=full_qual_name(pipeline),
                    artifact_global_url=None,
                    artifact_local_path=None,
                    params=dict(__init__=dict(task='token-classification',
                                              model=None,
                                              config=None,
                                              tokenizer=None,
                                              feature_extractor=None,
                                              framework=None,
                                              revision=None,
                                              use_fast=True,
                                              use_auth_token=None,
                                              model_kwargs={}))))[0])
        print(
            PreTrainedHuggingFaceTokenClassifier.objects.update_or_create(
                name='PreTrained-HuggingFace-Named-Entity-Recognizer',
                defaults=dict(
                    loader_module_and_qualname=full_qual_name(pipeline),
                    artifact_global_url=None,
                    artifact_local_path=None,
                    params=dict(__init__=dict(task='ner',
                                              model=None,
                                              config=None,
                                              tokenizer=None,
                                              feature_extractor=None,
                                              framework=None,
                                              revision=None,
                                              use_fast=True,
                                              use_auth_token=None,
                                              model_kwargs={}))))[0])
    except Exception as err:   # pylint: disable=broad-except
        print(f'{PreTrainedHuggingFaceTokenClassifier.__name__} model '
              f'not set up: {err}')

    for translation_model_name in ('facebook/m2m100_418M',
                                   'facebook/m2m100_1.2B',
                                   'facebook/mbart-large-50-one-to-many-mmt',
                                   'facebook/mbart-large-50-many-to-many-mmt'):
        try:
            print(
                PreTrainedHuggingFaceTranslator.objects.update_or_create(
                    name=('PreTrained-HuggingFace-Translator-' +
                          translation_model_name.replace('/', '-')),
                    defaults=dict(
                        loader_module_and_qualname=full_qual_name(pipeline),
                        artifact_global_url=None,
                        artifact_local_path=None,
                        params=dict(__init__=dict(task='translation',
                                                  model=translation_model_name,
                                                  config=None,
                                                  tokenizer=None,
                                                  feature_extractor=None,
                                                  framework=None,
                                                  revision=None,
                                                  use_fast=True,
                                                  use_auth_token=None,
                                                  model_kwargs={}))))[0])
        except Exception as err:   # pylint: disable=broad-except
            print(f'{PreTrainedHuggingFaceTranslator.__name__} model '
                  f'not set up: {err}')

    try:
        print(
            PreTrainedHuggingFaceZeroShotClassifier.objects.update_or_create(
                name='PreTrained-HuggingFace-Zero-Shot-Classifier',
                defaults=dict(
                    loader_module_and_qualname=full_qual_name(pipeline),
                    artifact_global_url=None,
                    artifact_local_path=None,
                    params=dict(__init__=dict(task='zero-shot-classification',
                                              model=None,
                                              config=None,
                                              tokenizer=None,
                                              feature_extractor=None,
                                              framework=None,
                                              revision=None,
                                              use_fast=True,
                                              use_auth_token=None,
                                              model_kwargs={}))))[0])
    except Exception as err:   # pylint: disable=broad-except
        print(f'{PreTrainedHuggingFaceZeroShotClassifier.__name__} model '
              f'not set up: {err}')



"""Djaizz Torch Deep Learning Model class."""


from sys import version_info

from torch.serialization import load

from ...base import _AIModelWithArtifactFilesABC

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


__all__: Sequence[str] = ('TorchModel',)


class TorchModel(_AIModelWithArtifactFilesABC):
    # pylint: disable=abstract-method
    """Djaizz Torch Deep Learning Model class."""

    class Meta(_AIModelWithArtifactFilesABC.Meta):
        # pylint: disable=too-few-public-methods
        """Django Model Class Metadata."""

        abstract = True

    def load(self) -> None:
        """Load the Model's native object from it artifact file."""
        if not self.native_obj:
            self.native_obj = load(f=self.artifact_local_path)



"""Djaizz Data Admin."""


from django.contrib.admin.decorators import register
from django.contrib.admin.options import ModelAdmin
from django.contrib.admin.sites import site

from silk.profiling.profiler import silk_profile

from djaizz.data.models import (
    DataSchema,
)


@register(DataSchema, site=site)
class DataSchemaAdmin(ModelAdmin):
    """DataSchema model admin."""

    show_full_result_count = False

    @silk_profile(name=f'{__module__}: {DataSchema._meta.verbose_name}')
    def changeform_view(self, *args, **kwargs):
        """Silk-profiled change-form view."""
        return super().changeform_view(*args, **kwargs)

    @silk_profile(name=f'{__module__}: {DataSchema._meta.verbose_name_plural}')
    def changelist_view(self, *args, **kwargs):
        """Silk-profiled change-list view."""
        return super().changelist_view(*args, **kwargs)



# pylint: disable=invalid-name


"""Djaizz In-Database JSON DataSet class."""


import json.decoder

import django.core.serializers.json
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    """Create Djaizz In-Database JSON DataSet class."""

    dependencies = (('AIData', '0001_DataSchema_DataSet'),)

    operations = (
        migrations.CreateModel(
            name='InDBJSONDataSet',

            fields=(
                ('dataset_ptr',
                 models.OneToOneField(
                    auto_created=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    parent_link=True,
                    primary_key=True,
                    related_name='in_db_json_data_sets',
                    serialize=False,
                    to='AIData.dataset')),

                ('in_db_json',
                 models.JSONField(
                    blank=True,
                    decoder=json.decoder.JSONDecoder,
                    default=None,
                    encoder=django.core.serializers.json.DjangoJSONEncoder,
                    help_text='In-Database JSON Data Content',
                    null=True,
                    verbose_name='In-Database JSON Data Content'))
            ),

            options={
                'verbose_name': 'In-Database JSON Data Set',
                'verbose_name_plural': 'In-Database JSON Data Sets',
                'db_table': 'AIData_InDBJSONDataSet',
                'abstract': False,
                'default_related_name': 'in_db_json_data_sets',
                'base_manager_name': 'objects'
            },

            bases=('AIData.dataset',)
        ),
    )



"""Djaizz base AIModel class."""


from abc import abstractmethod
from json.decoder import JSONDecoder
from sys import version_info
from typing import Any, Optional

from django.core.serializers.json import DjangoJSONEncoder
from django.db.models.fields import CharField
from django.db.models.fields.json import JSONField
from django.utils.functional import classproperty

from polymorphic.base import PolymorphicModelBase
from polymorphic.models import PolymorphicModel

from django_plotly_dash import DjangoDash

# pylint: disable=ungrouped-imports
try:
    from dash import dcc, html
except ImportError:
    import dash_core_components as dcc
    import dash_html_components as html
from dash.dependencies import Input, Output

import dash_bootstrap_components as dbc

from gradio.interface import Interface
from gradio.outputs import JSON as JSONOutput   # noqa: N811

from djutil.models import _ModelWithUUIDPKAndOptionalUniqueNameAndTimestampsABC

from djaizz.model.apps import DjaizzModelModuleConfig
from djaizz.util import PGSQL_IDENTIFIER_MAX_LEN, full_qual_name

if version_info >= (3, 9):
    from collections.abc import Generator, Sequence
else:
    from typing import Generator, Sequence  # pylint: disable=ungrouped-imports


__all__: Sequence[str] = 'AIModel', '_AIModelWithArtifactFilesABC'


class AIModel(PolymorphicModel,
              _ModelWithUUIDPKAndOptionalUniqueNameAndTimestampsABC):
    """Djaizz base AIModel class."""

    params: JSONField = \
        JSONField(
            verbose_name='Model Parameters',
            help_text='Model Parameters',

            encoder=DjangoJSONEncoder,
            decoder=JSONDecoder,

            null=True,
            blank=True,
            choices=None,
            db_column=None,
            db_index=False,
            db_tablespace=None,
            default=None,
            editable=True,
            # error_messages={},
            primary_key=False,
            unique=False,
            unique_for_date=None, unique_for_month=None, unique_for_year=None,
            # validators=None
        )

    class Meta(_ModelWithUUIDPKAndOptionalUniqueNameAndTimestampsABC.Meta):
        # pylint: disable=too-few-public-methods
        """Django Model Class Metadata."""

        verbose_name: str = 'AI Model'
        verbose_name_plural: str = 'AI Models'

        db_table: str = (f'{DjaizzModelModuleConfig.label}_'
                         f"{__qualname__.split(sep='.', maxsplit=1)[0]}")
        assert len(db_table) <= PGSQL_IDENTIFIER_MAX_LEN, \
            ValueError(f'*** "{db_table}" DB TABLE NAME TOO LONG ***')

        default_related_name: str = 'ai_models'

    @abstractmethod
    def fit(self, **kwargs) -> Any:
        """Fit."""
        raise NotImplementedError

    @abstractmethod
    def predict(self, **kwargs) -> Any:
        """Predict."""
        raise NotImplementedError

    @classproperty
    def _subclasses(cls) -> Generator[PolymorphicModelBase, None, None]:   # noqa: E501,N805
        # pylint: disable=no-self-argument
        """Generate AIModel class's subclasses."""
        for subclass in cls.__subclasses__():
            yield subclass
            yield from subclass._subclasses

    @classproperty
    def subclasses(cls) -> list[PolymorphicModelBase]:   # noqa: N805
        # pylint: disable=no-self-argument
        """List of AIModel class's subclasses."""
        return list(cls._subclasses)

    @classproperty
    def subclass_names(cls) -> list[str]:   # noqa: N805
        # pylint: disable=no-self-argument
        """List of AIModel class's subclasses' names."""
        return [s.__name__
                for s in cls._subclasses   # pylint: disable=not-an-iterable
                ]

    @classproperty
    def subclasses_by_name(cls) -> dict[str, PolymorphicModelBase]:   # noqa: E501,N805
        # pylint: disable=no-self-argument
        """Return AIModel class's subclass-by-name dictionary.

        Dictionary mapping from AIModel class's subclasses' names
        to such subclasses.
        """
        return {s.__name__: s
                for s in cls._subclasses   # pylint: disable=not-an-iterable
                }

    @classproperty
    def subclass_full_qual_names(cls) -> list[str]:   # noqa: N805
        # pylint: disable=no-self-argument
        """List of AIModel class's subclasses' fully-qualified names."""
        return [full_qual_name(s)
                for s in cls._subclasses   # pylint: disable=not-an-iterable
                ]

    @classproperty
    def subclasses_by_full_qual_name(cls) -> dict[str, PolymorphicModelBase]:   # noqa: E501,N805
        # pylint: disable=no-self-argument
        """Return AIModel class's subclass-by-fully-qualified name dict.

        Dictionary mapping from AIModel class's subclasses'
        fully-qualified names to such subclasses.
        """
        return {full_qual_name(s): s
                for s in cls._subclasses   # pylint: disable=not-an-iterable
                }

    @classproperty
    def dash_ui(cls) -> DjangoDash:   # noqa: N805
        # pylint: disable=no-self-argument
        """Return the AIModel class's Dash Interface."""
        app = DjangoDash(name=cls.__name__,
                         serve_locally=False,
                         add_bootstrap_links=True,
                         suppress_callback_exceptions=False,
                         external_stylesheets=[dbc.themes.SUPERHERO],
                         external_scripts=None)

        app.layout = html.Div(children=[
            html.H1(
                'Override the `dash_ui` classproperty to create a Dash UI for '
                f'{cls._meta.verbose_name_plural}'),

            dcc.Dropdown(
                id='ai-model-dropdown',
                # (string; optional):
                # The ID of this component,
                # used to identify dash components in callbacks.
                # The ID needs to be unique across all of the app's components.

                className='dropdown',
                # (string; optional): className of the dropdown element.

                clearable=True,
                # (boolean; default True):
                # Whether or not the dropdown is "clearable",
                # that is, whether or not a small "x" appears on the right
                # of the dropdown that removes the selected value.

                disabled=False,
                # (boolean; default False):
                # If True, this dropdown is disabled and the selection
                # cannot be changed.

                loading_state=None,
                # (dict; optional):
                # Object that holds the loading state object
                # coming from dash-renderer.
                # loading_state is a dict with keys:
                # - component_name (string; optional):
                #     Holds the name of the component that is loading.
                # - is_loading (boolean; optional):
                #     Determines if the component is loading or not.
                # - prop_name (string; optional):
                #     Holds which property is loading.

                multi=False,
                # (boolean; default False):
                # If True, the user can select multiple values.

                optionHeight=48,
                # (number; default 35): height of each option.
                # Can be increased when label lengths would wrap around.

                options=[dict(disabled=False, label=i, title=i, value=i)
                         for i in cls.names_or_uuids],
                # (list of dicts; optional): An array of options
                # {label: [string|number], value: [string|number]},
                # an optional disabled field can be used for each option.
                # options is a list of dicts with keys:
                # - disabled (boolean; optional):
                #     If True, this option is disabled and cannot be selected.
                # - label (string | number; required): The dropdown's label.
                # - title (string; optional):
                #     The HTML 'title' attribute for the option.
                #     Allows for information on hover.
                #     For more information on this attribute, see
                # developer.mozilla.org/en-US/docs/Web/HTML/Global_attributes/title.
                # - value (string | number; required):
                #     The value of the dropdown.
                #     This value corresponds to the items specified
                #     in the value property.

                persisted_props=['value'],
                # (list of values equal to: 'value'; default ['value']):
                # Properties whose user interactions will persist
                # after refreshing the component or the page. Since only value
                # is allowed this prop can normally be ignored.

                persistence=True,
                # (boolean | string | number; optional):
                # Used to allow user interactions in this component
                # to be persisted when the component - or the page -
                # is refreshed. If persisted is truthy and hasn't changed
                # from its previous value, a value that the user has changed
                # while using the app will keep that change,
                # as long as the new value also matches what was given
                # originally. Used in conjunction with persistence_type.

                persistence_type='session',
                # (a value equal to: 'local', 'session' or 'memory';
                # default 'local'):
                # Where persisted user changes will be stored:
                # - memory: only kept in memory, reset on page refresh.
                # - local: window.localStorage,
                #          data is kept after the browser quit.
                # - session: window.sessionStorage,
                #            data is cleared once the browser quit.

                placeholder='AI Model Name or UUID',
                # (string; optional):
                # The grey, default text shown when no option is selected.

                search_value=None,
                # (string; optional):
                # The value typed in the DropDown for searching.

                searchable=True,
                # (boolean; default True):
                # Whether to enable the searching feature or not.

                style=None,
                # (dict; optional):
                # Defines CSS styles which will override styles previously set.

                value=None,
                # (string | number | list of strings | numbers; optional):
                # The value of the input.
                # If multi is False (the default) then value is
                # just a string that corresponds to the values provided
                # in the options property.
                # If multi is True,
                # then multiple values can be selected at once,
                # and value is an array of items with values corresponding
                # to those in the options prop.
            ),

            html.Div(id='ai-model-dropdown-output-container'),
            ],   # noqa: E123
            # (list of or a singular dash component, string or number;
            # optional): The children of this component.

            # id=...,
            # (string; optional): The ID of this component,
            # used to identify dash components in callbacks.
            # The ID needs to be unique across all of the components in an app.

            accessKey=None,
            # (string; optional):
            # Keyboard shortcut to activate or add focus to the element.

            # aria-* (string; optional): A wildcard aria attribute.

            className=None,
            # (string; optional):
            # Often used with CSS to style elements with common properties.

            contentEditable=True,
            # (string; optional):
            # Indicates whether the element's content is editable.

            contextMenu=None,
            # (string; optional):
            # Defines the ID of a <menu> element
            # which will serve as the element's context menu.

            # data-* (string; optional): A wildcard data attribute.

            dir='ltr',
            # (string; optional): Defines the text direction.
            # Allowed values are ltr (Left-To-Right) or rtl (Right-To-Left).

            draggable=True,
            # (string; optional): Defines whether the element can be dragged.

            hidden=False,
            # (a value equal to: 'hidden' or 'HIDDEN' | boolean; optional):
            # Prevents rendering of given element,
            # while keeping child elements, e.g. script elements, active.

            key=None,
            # (string; optional): A unique identifier for the component,
            # used to improve performance by React.js while rendering
            # components. See https://reactjs.org/docs/lists-and-keys.html
            # for more info.

            lang='en',
            # (string; optional): Defines the language used in the element.

            loading_state=None,
            # (dict; optional):
            # Object that holds the loading state object
            # coming from dash-renderer.
            # loading_state is a dict with keys:
            # - component_name (string; optional):
            #     Holds the name of the component that is loading.
            # - is_loading (boolean; optional):
            #     Determines if the component is loading or not.
            # - prop_name (string; optional):
            # Holds which property is loading.

            n_clicks=0,
            # (number; default 0):
            # An integer that represents the number of times that
            # this element has been clicked on.

            n_clicks_timestamp=-1,
            # (number; default -1):
            # An integer that represents the time (in ms since 1970)
            # at which n_clicks changed.
            # This can be used to tell which button was changed most recently.

            role=None,
            # (string; optional): The ARIA role attribute.

            spellCheck=True,
            # (string; optional):
            # Indicates whether spell checking is allowed for the element.

            style={'height': '88%', 'width': '88%'},
            # (dict; optional):
            # Defines CSS styles which will override styles previously set.

            tabIndex=None,
            # (string; optional):
            # Overrides the browser's default tab order and
            # follows the one specified instead.

            title=None
            # (string; optional):
            # Text to be displayed in a tooltip when hovering over the element.
        )

        @app.callback(Output(component_id='ai-model-dropdown-output-container',
                             component_property='children'),
                      Input(component_id='ai-model-dropdown',
                            component_property='value'))
        def update_ai_model_dropdown_output(value):
            return f'{value} selected' if value else None

        return app

    @classproperty
    def gradio_ui(cls) -> Interface:   # noqa: N805
        # pylint: disable=no-self-argument
        """Return the AIModel class's Gradio Interface."""
        return Interface(
            fn=cls.predict,
            # (Callable) - the function to wrap an interface around.

            inputs=[],
            # (Union[str, list[Union[str, InputComponent]]]) -
            # a single Gradio input component,
            # or list of Gradio input components.
            # Components can either be passed as instantiated objects,
            # or referred to by their string shortcuts.
            # The number of input components should match
            # the number of parameters in fn.

            outputs=JSONOutput(label='AI Model Output'),
            # (Union[str, list[Union[str, OutputComponent]]]) -
            # a single Gradio output component,
            # or list of Gradio output components.
            # Components can either be passed as instantiated objects,
            # or referred to by their string shortcuts.
            # The number of output components should match
            # the number of values returned by fn.

            verbose=True,
            # (bool) - whether to print detailed information during launch.

            examples=None,
            # (Union[list[list[Any]], str]) - sample inputs for the function;
            # if provided, appears below the UI components and can be used
            # to populate the interface.
            # Should be nested list, in which the outer list consists of
            # samples and each inner list consists of an input
            # corresponding to each input component.
            # A string path to a directory of examples can also be provided.
            # If there are multiple input components and a directory
            # is provided, a log.csv file must be present in the directory
            # to link corresponding inputs.

            examples_per_page=10,
            # (int) - If examples are provided, how many to display per page.

            live=False,
            # (bool) - should the interface automatically reload on change?

            layout='unaligned',
            # (str) - Layout of input and output panels.
            # - "horizontal" arranges them as two columns of equal height,
            # - "unaligned" arranges them as two columns of unequal height, and
            # - "vertical" arranges them vertically.

            show_input=True,
            show_output=True,

            capture_session=False,
            # (bool) - if True, captures the default graph and session
            # (needed for Tensorflow 1.x)

            interpretation='default',
            # (Union[Callable, str]) - function that provides interpretation
            # explaining prediction output.
            # Pass "default" to use built-in interpreter.

            num_shap=2.0,
            # (float) - a multiplier that determines how many examples
            # are computed for shap-based interpretation.
            # Increasing this value will increase shap runtime,
            # but improve results.

            theme='default',
            # (str) - Theme to use - one of
            # - "default",
            # - "huggingface",
            # - "grass",
            # - "peach".
            # Add "dark" prefix, e.g. "darkpeach" or "darkdefault"
            # for darktheme.

            repeat_outputs_per_model=True,

            title=cls._meta.verbose_name,
            # (str) - a title for the interface;
            # if provided, appears above the input and output components.

            description=('Cverride the `gradio_ui` classproperty to create a '
                         f'Gradio UI for {cls._meta.verbose_name_plural}'),
            # (str) - a description for the interface;
            # if provided, appears above the input and output components.

            article=None,
            # (str) - an expanded article explaining the interface;
            # if provided, appears below the input and output components.
            # Accepts Markdown and HTML content.

            thumbnail=None,
            # (str) - path to image or src to use as display picture for models
            # listed in gradio.app/hub

            css=None,
            # (str) - custom css or path to custom css file
            # to use with interface.

            server_port=None,
            # (int) - will start gradio app on this port (if available)

            # server_name=networking.LOCALHOST_NAME,
            # (str) - to make app accessible on local network set to "0.0.0.0".

            height=500,
            width=900,

            allow_screenshot=True,
            # (bool) - if False, users will not see a button
            # to take a screenshot of the interface.

            allow_flagging=False,
            # (bool) - if False, users will not see a button
            # to flag an input and output.

            flagging_options=None,
            # (list[str]) - if not None, provides options a user must select
            # when flagging.

            encrypt=False,
            # (bool) - If True, flagged data will be encrypted
            # by key provided by creator at launch

            show_tips=False,
            # (bool) - if True, will occasionally show tips
            # about new Gradio features

            flagging_dir='flagged',
            # (str) - what to name the dir where flagged data is stored.

            analytics_enabled=True,

            enable_queue=False,
            # (bool) - if True, inference requests will be served through
            # a queue instead of with parallel threads.
            # Required for longer inference times (> 1min) to prevent timeout.
        )


class _AIModelWithArtifactFilesABC(AIModel):
    artifact_global_url: CharField = \
        CharField(
            verbose_name='Model Artifact Global URL',
            help_text='Model Artifact Global URL',

            max_length=255,

            null=True,
            blank=True,
            choices=None,
            db_column=None,
            db_index=True,
            db_tablespace=None,
            default=None,
            editable=True,
            # error_messages={},
            primary_key=False,
            unique=True,
            unique_for_date=None, unique_for_month=None, unique_for_year=None,
            # validators=None
        )

    artifact_local_path: CharField = \
        CharField(
            verbose_name='Model Artifact Local Path',
            help_text='Model Artifact Local Path',

            max_length=255,

            null=True,
            blank=True,
            choices=None,
            db_column=None,
            db_index=True,
            db_tablespace=None,
            default=None,
            editable=True,
            # error_messages={},
            primary_key=False,
            unique=False,
            unique_for_date=None, unique_for_month=None, unique_for_year=None,
            # validators=None
        )

    native_obj: Optional[Any] = None

    class Meta(AIModel.Meta):   # pylint: disable=too-few-public-methods
        """Django Model Class Metadata."""

        abstract = True

    @abstractmethod
    def load(self) -> None:
        """Load the Model's native object from its artifact file(s)."""
        raise NotImplementedError

    def unload(self) -> None:
        """Unload the Model's native object to free up memory."""
        self.native_obj = None



"""Djaizz Run-Command Utilities."""


from contextlib import AbstractContextManager
import os
from pathlib import Path
import shutil
from typing import Optional


from ..git import _GIT_HASH_FILE_NAME, get_git_repo_head_commit_hash


_SERVER_FILES_DIR_PATH: Path = Path(__file__).parent / '_server_files'


class ManagePyHandling(AbstractContextManager):
    """Handle `manage.py` file."""

    MANAGE_PY_FILE_NAME: str = 'manage.py'
    MANAGE_PY_FILE_SRC_PATH: Path = _SERVER_FILES_DIR_PATH / MANAGE_PY_FILE_NAME   # noqa: E501

    def __init__(self):
        """Initialize context manager."""
        self.manage_py_missing: bool = not Path(self.MANAGE_PY_FILE_NAME).exists()   # noqa: E501

    def __enter__(self):
        """Add `manage.py` file, if applicable."""
        if self.manage_py_missing:
            shutil.copyfile(src=self.MANAGE_PY_FILE_SRC_PATH,
                            dst=self.MANAGE_PY_FILE_NAME,
                            follow_symlinks=False)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Remove `manage.py` file, if applicable."""
        if self.manage_py_missing:
            os.remove(path=self.MANAGE_PY_FILE_NAME)


class SGIFilesHandling(AbstractContextManager):
    """Handle `asgi.py`/`wsgi.py` file & Procfile."""

    ASGI_PY_FILE_NAME: str = 'asgi.py'
    ASGI_PY_FILE_SRC_PATH: Path = _SERVER_FILES_DIR_PATH / ASGI_PY_FILE_NAME

    WSGI_PY_FILE_NAME: str = 'wsgi.py'
    WSGI_PY_FILE_SRC_PATH: Path = _SERVER_FILES_DIR_PATH / WSGI_PY_FILE_NAME

    PROCFILE_NAME: str = 'Procfile'

    def __init__(self, asgi: Optional[str] = None):
        """Initialize context manager."""
        self.asgi: Optional[str] = asgi

        if self.asgi:
            self.asgi_py_missing: bool = not Path(self.ASGI_PY_FILE_NAME).exists()   # noqa: E501
            self.procfile_missing: bool = not Path(self.PROCFILE_NAME).exists()

        else:
            self.wsgi_py_missing: bool = not Path(self.WSGI_PY_FILE_NAME).exists()   # noqa: E501

    def __enter__(self):
        """Add `asgi.py`/`wsgi.py` file & Procfile, if applicable."""
        if self.asgi:
            if self.asgi_py_missing:
                shutil.copyfile(src=self.ASGI_PY_FILE_SRC_PATH,
                                dst=self.ASGI_PY_FILE_NAME,
                                follow_symlinks=False)

            if self.procfile_missing:
                shutil.copyfile(
                    src=(_SERVER_FILES_DIR_PATH /
                         f'{self.PROCFILE_NAME}.{self.asgi.capitalize()}'),
                    dst=self.PROCFILE_NAME,
                    follow_symlinks=False)

        elif self.wsgi_py_missing:
            shutil.copyfile(src=self.WSGI_PY_FILE_SRC_PATH,
                            dst=self.WSGI_PY_FILE_NAME,
                            follow_symlinks=False)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Remove `asgi.py`/`wsgi.py` file & Procfile, if applicable."""
        if self.asgi:
            if self.asgi_py_missing:
                os.remove(path=self.ASGI_PY_FILE_NAME)

            if self.procfile_missing:
                os.remove(path=self.PROCFILE_NAME)

        elif self.wsgi_py_missing:
            os.remove(path=self.WSGI_PY_FILE_NAME)


class GitHashHandling(AbstractContextManager):
    """Handle `Procfile` file."""

    def __init__(self):
        """Initialize context manager."""
        self.git_hash: Optional[str] = get_git_repo_head_commit_hash()

    def __enter__(self):
        """Add `.git-hash` file if applicable."""
        if self.git_hash:
            assert not Path(_GIT_HASH_FILE_NAME).exists()

            with open(file=_GIT_HASH_FILE_NAME,
                      mode='wt',
                      buffering=-1,
                      encoding='utf-8',
                      errors='strict',
                      newline=None,
                      closefd=True,
                      opener=None) as opened_file:
                opened_file.write(self.git_hash)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Remove `.git-hash` file, if applicable."""
        if self.git_hash:
            os.remove(path=_GIT_HASH_FILE_NAME)


def run_cmd(command: str,
            asgi: Optional[str] = None):
    """Run a CLI command."""
    with ManagePyHandling(), SGIFilesHandling(asgi=asgi), GitHashHandling():
        os.system(command=command)



#!/usr/bin/env python3


"""Django administration/management script."""


# ref: django-configurations.readthedocs.io


import os
import sys

from configurations.management import execute_from_command_line


if __name__ == '__main__':
    os.environ.setdefault(key='DJANGO_SETTINGS_MODULE', value='settings')
    os.environ.setdefault(key='DJANGO_CONFIGURATION', value='Default')

    execute_from_command_line(argv=sys.argv)



"""Djaizz Git-related Utilities."""


from pathlib import Path
from sys import version_info
from typing import Optional

from git.exc import InvalidGitRepositoryError   # pylint: disable=import-error
from git.repo.base import Repo   # pylint: disable=import-error

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence   # pylint: disable=ungrouped-imports


__all__: Sequence[str] = ('get_git_repo_head_commit_hash',)


_GIT_HASH_FILE_NAME: str = '.git-hash'


def get_git_repo_head_commit_hash(path: Optional[str] = None) -> Optional[str]:
    """Get Git Repo Head Commit Hash."""
    try:
        repo: Repo = Repo(path=path,
                          search_parent_directories=True,
                          expand_vars=True)

        return repo.head.commit.hexsha

    except InvalidGitRepositoryError:
        if Path(_GIT_HASH_FILE_NAME).is_file():
            with open(file=_GIT_HASH_FILE_NAME,
                      mode='rt',
                      buffering=-1,
                      encoding='utf-8',
                      errors='strict',
                      newline=None,
                      closefd=True,
                      opener=None) as opened_file:
                return opened_file.read()

        else:
            return None



"""Djaizz Model Module Config."""


from sys import version_info

from django.apps.config import AppConfig

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


__all__: Sequence[str] = ('DjaizzModelModuleConfig',)


# docs.djangoproject.com/en/dev/ref/applications/#application-configuration
class DjaizzModelModuleConfig(AppConfig):
    """Djaizz Model Module Config."""

    # AppConfig.name
    # Full Python path to the application, e.g. 'django.contrib.admin'.
    # This attribute defines which application the configuration applies to.
    # It must be set in all AppConfig subclasses.
    # It must be unique across a Django project.
    name: str = 'djaizz.model'

    # AppConfig.label
    # Short name for the application, e.g. 'admin'
    # This attribute allows relabeling an application
    # when two applications have conflicting labels.
    # It defaults to the last component of name.
    # It should be a valid Python identifier.
    # It must be unique across a Django project.
    label: str = 'AIModel'

    # AppConfig.verbose_name
    # Human-readable name for the application, e.g. “Administration”.
    # This attribute defaults to label.title().
    verbose_name: str = 'Djaizz: Models'

    # AppConfig.path
    # Filesystem path to the application directory,
    # e.g. '/usr/lib/pythonX.Y/dist-packages/django/contrib/admin'.
    # In most cases, Django can automatically detect and set this,
    # but you can also provide an explicit override as a class attribute
    # on your AppConfig subclass.
    # In a few situations this is required; for instance
    # if the app package is a namespace package with multiple paths.
    # path: str = ...

    # AppConfig.default
    # Set this attribute to False
    # to prevent Django from selecting a configuration class automatically.
    # This is useful when apps.py defines only one AppConfig subclass
    # but you don’t want Django to use it by default.
    # Set this attribute to True
    # to tell Django to select a configuration class automatically.
    # This is useful when apps.py defines more than one AppConfig subclass
    # and you want Django to use one of them by default.
    # By default, this attribute isn’t set.
    default: bool = True

    # AppConfig.default_auto_field
    # The implicit primary key type to add to models within this app.
    # You can use this to keep AutoField
    # as the primary key type for third party applications.
    # By default, this is the value of DEFAULT_AUTO_FIELD.
    default_auto_field: str = 'django.db.models.fields.AutoField'

    def ready(self) -> None:
        """Run tasks to initialize module."""
        # pylint: disable=import-outside-toplevel
        from djaizz.model.scripts import (
            setup_google_cloud_ai_svcs,
            setup_pretrained_huggingface_models,
            setup_pretrained_keras_models,
        )

        msg = ('SETTING UP CLOUD AI SERVICES & '
               'PRE-TRAINED DEEP LEARNING MODELS/PIPELINES...')
        print(f'\n{msg}\n')

        setup_google_cloud_ai_svcs.run()
        setup_pretrained_huggingface_models.run()
        setup_pretrained_keras_models.run()

        print(f'\n{msg} done!\n')



"""Djaizz Pre-Trained Hugging Face Audio Classifier Model class."""


from sys import version_info
from typing import Union

from django.utils.functional import classproperty

from gradio.interface import Interface
from gradio.inputs import (Audio as AudioInput,
                           Slider as SliderInput)
from gradio.outputs import Label as LabelOutput

import numpy

from djaizz.model.apps import DjaizzModelModuleConfig
from djaizz.util import PGSQL_IDENTIFIER_MAX_LEN

from .base import PreTrainedHuggingFaceTransformer

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence   # pylint: disable=ungrouped-imports


__all__: Sequence[str] = ('PreTrainedHuggingFaceAudioClassifier',)


AudioClassificationInputType = Union[numpy.ndarray, str]
AudioClassificationOutputType = dict[str, float]


class PreTrainedHuggingFaceAudioClassifier(PreTrainedHuggingFaceTransformer):
    # pylint: disable=abstract-method,too-many-ancestors
    """Djaizz Pre-Trained Hugging Face Audio Classifier Model class."""

    class Meta(PreTrainedHuggingFaceTransformer.Meta):
        # pylint: disable=too-few-public-methods
        """Django Model Class Metadata."""

        verbose_name: str = 'Pre-Trained Hugging Face Audio Classifier'
        verbose_name_plural: str = 'Pre-Trained Hugging Face Audio Classifiers'

        db_table: str = (f'{DjaizzModelModuleConfig.label}_'
                         f"{__qualname__.split(sep='.', maxsplit=1)[0]}")
        assert len(db_table) <= PGSQL_IDENTIFIER_MAX_LEN, \
            ValueError(f'*** "{db_table}" DB TABLE NAME TOO LONG ***')

        default_related_name: str = 'pretrained_hugging_face_audio_classifiers'

    def predict(self,
                audio_or_audios:
                    Union[AudioClassificationInputType,
                          Sequence[AudioClassificationInputType]],
                n_labels: int = 5) \
            -> Union[AudioClassificationOutputType,
                     list[AudioClassificationOutputType]]:
        # pylint: disable=arguments-differ
        """Classify Audio(s)."""
        single_audio: bool = isinstance(audio_or_audios, (numpy.ndarray, str))

        if not (single_audio or isinstance(audio_or_audios, list)):
            audio_or_audios: list[AudioClassificationInputType] = \
                list(audio_or_audios)

        self.load()

        output = self.native_obj(inputs=audio_or_audios, top_k=n_labels)

        return ({i['label']: i['score'] for i in output}
                if single_audio
                else [{i['label']: i['score'] for i in result}
                      for result in output])

    @classproperty
    def gradio_ui(cls) -> Interface:   # noqa: N805
        # pylint: disable=no-self-argument
        """Gradio Interface."""
        def _predict(self,
                     sampling_rate_and_double_channel_audio_array:
                     tuple[int, numpy.ndarray],
                     n_labels: int = 5) -> dict[str, float]:
            _sampling_rate, double_channel_audio_array = \
                sampling_rate_and_double_channel_audio_array

            return cls.predict(
                self,
                audio_or_audios=(double_channel_audio_array[:, 0]
                                 .astype(numpy.float32)),
                n_labels=n_labels)

        return Interface(
            fn=_predict,
            # (Callable) - the function to wrap an interface around.

            inputs=[AudioInput(source='upload',
                               type='numpy',
                               label='Audio to Classify',
                               optional=False),

                    SliderInput(minimum=3, maximum=10, step=1, default=5,
                                label='No. of Labels to Return')],
            # (Union[str, list[Union[str, InputComponent]]]) -
            # a single Gradio input component,
            # or list of Gradio input components.
            # Components can either be passed as instantiated objects,
            # or referred to by their string shortcuts.
            # The number of input components should match
            # the number of parameters in fn.

            outputs=LabelOutput(num_top_classes=10,
                                type='auto',
                                label='Audio Classification'),
            # (Union[str, list[Union[str, OutputComponent]]]) -
            # a single Gradio output component,
            # or list of Gradio output components.
            # Components can either be passed as instantiated objects,
            # or referred to by their string shortcuts.
            # The number of output components should match
            # the number of values returned by fn.

            verbose=True,
            # (bool) - whether to print detailed information during launch.

            examples=None,
            # (Union[list[list[Any]], str]) - sample inputs for the function;
            # if provided, appears below the UI components and can be used
            # to populate the interface.
            # Should be nested list, in which the outer list consists of
            # samples and each inner list consists of an input
            # corresponding to each input component.
            # A string path to a directory of examples can also be provided.
            # If there are multiple input components and a directory
            # is provided, a log.csv file must be present in the directory
            # to link corresponding inputs.

            examples_per_page=10,
            # (int) - If examples are provided, how many to display per page.

            live=False,
            # (bool) - should the interface automatically reload on change?

            layout='unaligned',
            # (str) - Layout of input and output panels.
            # - "horizontal" arranges them as two columns of equal height,
            # - "unaligned" arranges them as two columns of unequal height, and
            # - "vertical" arranges them vertically.

            show_input=True,
            show_output=True,

            capture_session=False,
            # (bool) - if True, captures the default graph and session
            # (needed for Tensorflow 1.x)

            interpretation='default',
            # (Union[Callable, str]) - function that provides interpretation
            # explaining prediction output.
            # Pass "default" to use built-in interpreter.

            num_shap=2.0,
            # (float) - a multiplier that determines how many examples
            # are computed for shap-based interpretation.
            # Increasing this value will increase shap runtime,
            # but improve results.

            theme='default',
            # (str) - Theme to use - one of
            # - "default",
            # - "compact",
            # - "huggingface", or
            # - "darkhuggingface".

            repeat_outputs_per_model=True,

            title=cls._meta.verbose_name,
            # (str) - a title for the interface;
            # if provided, appears above the input and output components.

            description='A pre-trained Hugging Face model to classify audio',
            # (str) - a description for the interface;
            # if provided, appears above the input and output components.

            article=None,
            # (str) - an expanded article explaining the interface;
            # if provided, appears below the input and output components.
            # Accepts Markdown and HTML content.

            thumbnail=None,
            # (str) - path to image or src to use as display picture for models
            # listed in gradio.app/hub

            css=None,
            # (str) - custom css or path to custom css file
            # to use with interface.

            server_port=None,
            # (int) - will start gradio app on this port (if available)

            # server_name=networking.LOCALHOST_NAME,
            # (str) - to make app accessible on local network set to "0.0.0.0".

            height=500,
            width=900,

            allow_screenshot=True,
            # (bool) - if False, users will not see a button
            # to take a screenshot of the interface.

            allow_flagging=False,
            # (bool) - if False, users will not see a button
            # to flag an input and output.

            flagging_options=None,
            # (list[str]) - if not None, provides options a user must select
            # when flagging.

            encrypt=False,
            # (bool) - If True, flagged data will be encrypted
            # by key provided by creator at launch

            show_tips=False,
            # (bool) - if True, will occasionally show tips
            # about new Gradio features

            flagging_dir='flagged',
            # (str) - what to name the dir where flagged data is stored.

            analytics_enabled=True,

            enable_queue=False,
            # (bool) - if True, inference requests will be served through
            # a queue instead of with parallel threads.
            # Required for longer inference times (> 1min) to prevent timeout.
        )



"""Djaizz NumPyArray class."""


from sys import version_info

from .json import _FileDataSetWithInDBJSONCacheABC

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


__all__: Sequence[str] = ('PandasDataFrame',)


class PandasDataFrame(_FileDataSetWithInDBJSONCacheABC):
    # pylint: disable=abstract-method,too-many-ancestors
    """Djaizz JSON DataSet class."""

    class Meta(_FileDataSetWithInDBJSONCacheABC.Meta):
        # pylint: disable=too-few-public-methods
        """Django Model Class Metadata."""

        abstract = True



# pylint: disable=invalid-name


"""Djaizz PreTrainedKerasImageNetClassifier model class."""


from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    """Create Djaizz PreTrainedKerasImageNetClassifier model class."""

    dependencies = (('AIModel', '0001_AIModel'),)

    operations = (
        migrations.CreateModel(
            name='PreTrainedKerasImageNetClassifier',

            fields=(
                ('aimodel_ptr',
                 models.OneToOneField(
                    auto_created=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    parent_link=True,
                    primary_key=True,
                    related_name='pretrained_keras_imagenet_classifiers',
                    serialize=False,
                    to='AIModel.aimodel')),

                ('artifact_global_url',
                 models.CharField(
                    blank=True,
                    db_index=True,
                    default=None,
                    help_text='Model Artifact Global URL',
                    max_length=255,
                    null=True,
                    unique=True,
                    verbose_name='Model Artifact Global URL')),

                ('artifact_local_path',
                 models.CharField(
                    blank=True,
                    db_index=True,
                    default=None,
                    help_text='Model Artifact Local Path',
                    max_length=255,
                    null=True,
                    verbose_name='Model Artifact Local Path')),

                ('loader_module_and_qualname',
                 models.CharField(
                    db_index=True,
                    default=None,
                    help_text=('Pre-Trained ML Model Loader '
                               '(module.dot.qualname)'),
                    max_length=255,
                    verbose_name=('Pre-Trained ML Model Loader '
                                  '(module.dot.qualname)'))),

                ('preprocessor_module_and_qualname',
                 models.CharField(
                    db_index=True,
                    default=None,
                    help_text='Preprocessor (module.dot.qualname)',
                    max_length=255,
                    verbose_name='Preprocessor (module.dot.qualname)'))
            ),

            options={
                'verbose_name': 'Pre-Trained Keras ImageNet Classifier',
                'verbose_name_plural':
                    'Pre-Trained Keras ImageNet Classifiers',
                'db_table': 'AIModel_PreTrainedKerasImageNetClassifier',
                'abstract': False,
                'default_related_name':
                    'pretrained_keras_imagenet_classifiers',
                'base_manager_name': 'objects'
            },

            bases=('AIModel.aimodel',)
        ),
    )



"""Djaizz Pre-Trained Hugging Face Transformer Model class."""


from sys import version_info

from djaizz.util import PGSQL_IDENTIFIER_MAX_LEN
from djaizz.model.apps import DjaizzModelModuleConfig

from ..base import _PreTrainedMLModelABC

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


__all__: Sequence[str] = ('PreTrainedHuggingFaceTransformer',)


class PreTrainedHuggingFaceTransformer(_PreTrainedMLModelABC):
    # pylint: disable=abstract-method
    """Djaizz Pre-Trained Hugging Face Transformer Model class."""

    class Meta(_PreTrainedMLModelABC.Meta):
        # pylint: disable=too-few-public-methods
        """Django Model Class Metadata."""

        verbose_name: str = 'Pre-Trained Hugging Face Transformer'
        verbose_name_plural: str = 'Pre-Trained Hugging Face Transformers'

        db_table: str = (f'{DjaizzModelModuleConfig.label}_'
                         f"{__qualname__.split(sep='.', maxsplit=1)[0]}")
        assert len(db_table) <= PGSQL_IDENTIFIER_MAX_LEN, \
            ValueError(f'*** "{db_table}" DB TABLE NAME TOO LONG ***')

        default_related_name: str = 'pretrained_hugging_face_transformers'



"""Migrations related to Django 4.0."""


# pylint: disable=invalid-name


from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    """Migrations related to Django 4.0."""

    dependencies = [('contenttypes', '0002_remove_content_type_name'),
                    ('AIModel', '0005_GoogleTranslate')]

    operations = [
        migrations.AlterField(
            model_name='aimodel',
            name='polymorphic_ctype',
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name='polymorphic_%(app_label)s.%(class)s_set+',
                to='contenttypes.contenttype')),

        migrations.AlterField(
            model_name='cloudaiservice',
            name='aimodel_ptr',
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to='AIModel.aimodel')),

        migrations.AlterField(
            model_name='googletranslate',
            name='cloudaiservice_ptr',
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to='AIModel.cloudaiservice')),

        migrations.AlterField(
            model_name='pretrainedhuggingfaceaudioclassifier',
            name='pretrainedhuggingfacetransformer_ptr',
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to='AIModel.pretrainedhuggingfacetransformer')),

        migrations.AlterField(
            model_name='pretrainedhuggingfaceimageclassifier',
            name='pretrainedhuggingfacetransformer_ptr',
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to='AIModel.pretrainedhuggingfacetransformer')),

        migrations.AlterField(
            model_name='pretrainedhuggingfacemaskfiller',
            name='pretrainedhuggingfacetransformer_ptr',
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to='AIModel.pretrainedhuggingfacetransformer')),

        migrations.AlterField(
            model_name='pretrainedhuggingfaceobjectdetector',
            name='pretrainedhuggingfacetransformer_ptr',
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to='AIModel.pretrainedhuggingfacetransformer')),

        migrations.AlterField(
            model_name='pretrainedhuggingfacequestionanswerer',
            name='pretrainedhuggingfacetransformer_ptr',
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to='AIModel.pretrainedhuggingfacetransformer')),

        migrations.AlterField(
            model_name='pretrainedhuggingfacespeechrecognizer',
            name='pretrainedhuggingfacetransformer_ptr',
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to='AIModel.pretrainedhuggingfacetransformer')),

        migrations.AlterField(
            model_name='pretrainedhuggingfacetablequestionanswerer',
            name='pretrainedhuggingfacetransformer_ptr',
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to='AIModel.pretrainedhuggingfacetransformer')),

        migrations.AlterField(
            model_name='pretrainedhuggingfacetext2textgenerator',
            name='pretrainedhuggingfacetransformer_ptr',
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to='AIModel.pretrainedhuggingfacetransformer')),

        migrations.AlterField(
            model_name='pretrainedhuggingfacetextclassifier',
            name='pretrainedhuggingfacetransformer_ptr',
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to='AIModel.pretrainedhuggingfacetransformer')),

        migrations.AlterField(
            model_name='pretrainedhuggingfacetextgenerator',
            name='pretrainedhuggingfacetransformer_ptr',
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to='AIModel.pretrainedhuggingfacetransformer')),

        migrations.AlterField(
            model_name='pretrainedhuggingfacetextsummarizer',
            name='pretrainedhuggingfacetransformer_ptr',
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to='AIModel.pretrainedhuggingfacetransformer')),

        migrations.AlterField(
            model_name='pretrainedhuggingfacetokenclassifier',
            name='pretrainedhuggingfacetransformer_ptr',
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to='AIModel.pretrainedhuggingfacetransformer')),

        migrations.AlterField(
            model_name='pretrainedhuggingfacetransformer',
            name='aimodel_ptr',
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to='AIModel.aimodel')),

        migrations.AlterField(
            model_name='pretrainedhuggingfacetranslator',
            name='pretrainedhuggingfacetransformer_ptr',
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to='AIModel.pretrainedhuggingfacetransformer')),

        migrations.AlterField(
            model_name='pretrainedhuggingfacezeroshotclassifier',
            name='pretrainedhuggingfacetransformer_ptr',
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to='AIModel.pretrainedhuggingfacetransformer')),

        migrations.AlterField(
            model_name='pretrainedkerasimagenetclassifier',
            name='aimodel_ptr',
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to='AIModel.aimodel'))
    ]



"""Djaizz Model URLs."""


from functools import partial
from sys import version_info

from django.urls.conf import path
from django.urls.resolvers import URLPattern

from djaizz.model.views import model_ui

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


__all__: Sequence[str] = ('urlpatterns',)


urlpatterns: Sequence[URLPattern] = (
    path(route='<str:model_class_or_instance_name_or_uuid>/dash/',
         view=partial(model_ui, ui_type='dash'),
         kwargs=None, name=None),

    path(route='<str:model_class_or_instance_name_or_uuid>/gradio/',
         view=partial(model_ui, ui_type='gradio'),
         kwargs=None, name=None),
)



"""Set up pre-trained Keras models."""


# pylint: disable=import-error,no-name-in-module,ungrouped-imports
from tensorflow.keras.applications.densenet import (
    DenseNet121, DenseNet169, DenseNet201,
    preprocess_input as densenet_preprocess_input)
from keras.applications.densenet import (
    DENSENET121_WEIGHT_PATH as DENSENET121_URL,
    DENSENET169_WEIGHT_PATH as DENSENET169_URL,
    DENSENET201_WEIGHT_PATH as DENSENET201_URL,
)
from tensorflow.keras.applications.efficientnet import (
    EfficientNetB0,
    EfficientNetB1,
    EfficientNetB2,
    EfficientNetB3,
    EfficientNetB4,
    EfficientNetB5,
    EfficientNetB6,
    EfficientNetB7,
    preprocess_input as efficientnet_preprocess_input)
from keras.applications.efficientnet import (
    BASE_WEIGHTS_PATH as EFFICIENTNET_BASE_URL,
)
from tensorflow.keras.applications.inception_v3 import (
    InceptionV3,
    preprocess_input as inception_preprocess_input)
from keras.applications.inception_v3 import (
    WEIGHTS_PATH as INCEPTION_URL,
)
from tensorflow.keras.applications.inception_resnet_v2 import (
    InceptionResNetV2,
    preprocess_input as inception_resnet_preprocess_input)
from keras.applications.inception_resnet_v2 import (
    BASE_WEIGHT_URL as INCEPTION_RESNET_BASE_URL,
)
from tensorflow.keras.applications.mobilenet import (
    MobileNet,
    preprocess_input as mobilenet_v1_preprocess_input)
from keras.applications.mobilenet import (
    BASE_WEIGHT_PATH as MOBILENET_V1_BASE_URL,
)
from tensorflow.keras.applications.mobilenet_v2 import (
    MobileNetV2,
    preprocess_input as mobilenet_v2_preprocess_input)
from keras.applications.mobilenet_v2 import (
    BASE_WEIGHT_PATH as MOBILENET_V2_BASE_URL,
)
from tensorflow.keras.applications.mobilenet_v3 import (
    preprocess_input as mobilenet_v3_preprocess_input)
from keras.applications.mobilenet_v3 import (
    MobileNetV3Large, MobileNetV3Small,
    BASE_WEIGHT_PATH as MOBILENET_V3_BASE_URL,
)
from tensorflow.keras.applications.nasnet import (
    NASNetLarge, NASNetMobile,
    preprocess_input as nasnet_preprocess_input)
from keras.applications.nasnet import (
    NASNET_LARGE_WEIGHT_PATH as NASNET_LARGE_URL,
    NASNET_MOBILE_WEIGHT_PATH as NASNET_MOBILE_URL,
)
from tensorflow.keras.applications.resnet import (
    ResNet50, ResNet101, ResNet152,
    preprocess_input as resnet_v1_preprocess_input)
from keras.applications.resnet import (
    BASE_WEIGHTS_PATH as RESNET_BASE_URL,
)
from tensorflow.keras.applications.resnet_v2 import (
    ResNet50V2, ResNet101V2, ResNet152V2,
    preprocess_input as resnet_v2_preprocess_input)
from tensorflow.keras.applications.vgg16 import (
    VGG16,
    preprocess_input as vgg_preprocess_input)
from keras.applications.vgg16 import (
    WEIGHTS_PATH as VGG16_URL,
)
from tensorflow.keras.applications.vgg19 import VGG19
from keras.applications.vgg19 import (
    WEIGHTS_PATH as VGG19_URL,
)
from tensorflow.keras.applications.xception import (
    Xception,
    preprocess_input as xception_preprocess_input)
from keras.applications.xception import (
    TF_WEIGHTS_PATH as XCEPTION_URL,
)

from djaizz.model.models.ml.keras import PreTrainedKerasImageNetClassifier
from djaizz.util import full_qual_name


MODEL_SPECS = (
    (DenseNet121, densenet_preprocess_input, 224,
     DENSENET121_URL,
     'densenet121_weights_tf_dim_ordering_tf_kernels.h5'),

    (DenseNet169, densenet_preprocess_input, 224,
     DENSENET169_URL,
     'densenet169_weights_tf_dim_ordering_tf_kernels.h5'),

    (DenseNet201, densenet_preprocess_input, 224,
     DENSENET201_URL,
     'densenet201_weights_tf_dim_ordering_tf_kernels.h5'),

    (EfficientNetB0, efficientnet_preprocess_input, 224,
     f'{EFFICIENTNET_BASE_URL}efficientnetb0.h5',
     'efficientnetb0.h5'),

    (EfficientNetB1, efficientnet_preprocess_input, 240,
     f'{EFFICIENTNET_BASE_URL}efficientnetb1.h5',
     'efficientnetb1.h5'),

    (EfficientNetB2, efficientnet_preprocess_input, 260,
     f'{EFFICIENTNET_BASE_URL}efficientnetb2.h5',
     'efficientnetb2.h5'),

    (EfficientNetB3, efficientnet_preprocess_input, 300,
     f'{EFFICIENTNET_BASE_URL}efficientnetb3.h5',
     'efficientnetb3.h5'),

    (EfficientNetB4, efficientnet_preprocess_input, 380,
     f'{EFFICIENTNET_BASE_URL}efficientnetb4.h5',
     'efficientnetb4.h5'),

    (EfficientNetB5, efficientnet_preprocess_input, 456,
     f'{EFFICIENTNET_BASE_URL}efficientnetb5.h5',
     'efficientnetb5.h5'),

    (EfficientNetB6, efficientnet_preprocess_input, 528,
     f'{EFFICIENTNET_BASE_URL}efficientnetb6.h5',
     'efficientnetb6.h5'),

    (EfficientNetB7, efficientnet_preprocess_input, 600,
     f'{EFFICIENTNET_BASE_URL}efficientnetb7.h5',
     'efficientnetb7.h5'),

    (InceptionV3, inception_preprocess_input, 299,
     INCEPTION_URL,
     'inception_v3_weights_tf_dim_ordering_tf_kernels.h5'),

    (InceptionResNetV2, inception_resnet_preprocess_input, 299,
     f'{INCEPTION_RESNET_BASE_URL}'
     'inception_resnet_v2_weights_tf_dim_ordering_tf_kernels.h5',
     'inception_resnet_v2_weights_tf_dim_ordering_tf_kernels.h5'),

    (MobileNet, mobilenet_v1_preprocess_input, 224,
     f'{MOBILENET_V1_BASE_URL}mobilenet_1_0_224_tf.h5',
     'mobilenet_1_0_224_tf.h5'),

    (MobileNetV2, mobilenet_v2_preprocess_input, 224,
     f'{MOBILENET_V2_BASE_URL}'
     'mobilenet_v2_weights_tf_dim_ordering_tf_kernels_1.0_224.h5',
     'mobilenet_v2_weights_tf_dim_ordering_tf_kernels_1.0_224.h5'),

    (MobileNetV3Large, mobilenet_v3_preprocess_input, 224,
     f'{MOBILENET_V3_BASE_URL}weights_mobilenet_v3_large_224_1.0_float.h5',
     'weights_mobilenet_v3_large_224_1.0_float.h5'),

    (MobileNetV3Small, mobilenet_v3_preprocess_input, 224,
     f'{MOBILENET_V3_BASE_URL}weights_mobilenet_v3_small_224_1.0_float.h5',
     'weights_mobilenet_v3_small_224_1.0_float.h5'),

    (NASNetLarge, nasnet_preprocess_input, 331,
     NASNET_LARGE_URL,
     'nasnet_large.h5'),

    (NASNetMobile, nasnet_preprocess_input, 224,
     NASNET_MOBILE_URL,
     'nasnet_mobile.h5'),

    (ResNet50, resnet_v1_preprocess_input, 224,
     f'{RESNET_BASE_URL}resnet50_weights_tf_dim_ordering_tf_kernels.h5',
     'resnet50_weights_tf_dim_ordering_tf_kernels.h5'),

    (ResNet101, resnet_v1_preprocess_input, 224,
     f'{RESNET_BASE_URL}resnet101_weights_tf_dim_ordering_tf_kernels.h5',
     'resnet101_weights_tf_dim_ordering_tf_kernels.h5'),

    (ResNet152, resnet_v1_preprocess_input, 224,
     f'{RESNET_BASE_URL}resnet152_weights_tf_dim_ordering_tf_kernels.h5',
     'resnet152_weights_tf_dim_ordering_tf_kernels.h5'),

    (ResNet50V2, resnet_v2_preprocess_input, 224,
     f'{RESNET_BASE_URL}resnet50v2_weights_tf_dim_ordering_tf_kernels.h5',
     'resnet50v2_weights_tf_dim_ordering_tf_kernels.h5'),

    (ResNet101V2, resnet_v2_preprocess_input, 224,
     f'{RESNET_BASE_URL}resnet101v2_weights_tf_dim_ordering_tf_kernels.h5',
     'resnet101v2_weights_tf_dim_ordering_tf_kernels.h5'),

    (ResNet152V2, resnet_v2_preprocess_input, 224,
     f'{RESNET_BASE_URL}resnet152v2_weights_tf_dim_ordering_tf_kernels.h5',
     'resnet152v2_weights_tf_dim_ordering_tf_kernels.h5'),

    (VGG16, vgg_preprocess_input, 224,
     VGG16_URL,
     'vgg16_weights_tf_dim_ordering_tf_kernels.h5'),

    (VGG19, vgg_preprocess_input, 224,
     VGG19_URL,
     'vgg19_weights_tf_dim_ordering_tf_kernels.h5'),

    (Xception, xception_preprocess_input, 299,
     XCEPTION_URL,
     'xception_weights_tf_dim_ordering_tf_kernels.h5')
)


def run():
    """Run this script."""
    model_name_prefix = f'{PreTrainedKerasImageNetClassifier.__name__}-'

    for keras_loader, preproc, img_dim_size, global_url, local_file_name in \
            MODEL_SPECS:
        try:
            print(PreTrainedKerasImageNetClassifier.objects.update_or_create(
                name=model_name_prefix + keras_loader.__name__,
                defaults=dict(
                    loader_module_and_qualname=full_qual_name(keras_loader),
                    preprocessor_module_and_qualname=full_qual_name(preproc),
                    artifact_global_url=global_url,
                    artifact_local_path=f'~/.keras/models/{local_file_name}',
                    params=dict(img_dim_size=img_dim_size)))[0])

        except Exception as err:   # pylint: disable=broad-except
            print(f'{PreTrainedKerasImageNetClassifier.__name__} model '
                  f'not set up: {err}')



"""Djaizz Model Public API."""


from sys import version_info

from djaizz.model.models import (
    AIModel,

    CloudAIService,

    GoogleTranslate,

    SKLModel,

    KerasModel,
    PreTrainedKerasImageNetClassifier,

    TorchModel,

    PreTrainedHuggingFaceTransformer,
    PreTrainedHuggingFaceAudioClassifier,
    PreTrainedHuggingFaceImageClassifier,
    PreTrainedHuggingFaceMaskFiller,
    PreTrainedHuggingFaceObjectDetector,
    PreTrainedHuggingFaceQuestionAnswerer,
    PreTrainedHuggingFaceSpeechRecognizer,
    PreTrainedHuggingFaceTableQuestionAnswerer,
    PreTrainedHuggingFaceTextClassifier,
    PreTrainedHuggingFaceTextGenerator,
    PreTrainedHuggingFaceText2TextGenerator,
    PreTrainedHuggingFaceTextSummarizer,
    PreTrainedHuggingFaceTokenClassifier,
    PreTrainedHuggingFaceTranslator,
    PreTrainedHuggingFaceZeroShotClassifier,
)

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


__all__: Sequence[str] = (
    'AIModel',

    'CloudAIService',

    'GoogleTranslate',

    'SKLModel',

    'KerasModel',
    'PreTrainedKerasImageNetClassifier',

    'TorchModel',

    'PreTrainedHuggingFaceTransformer',
    'PreTrainedHuggingFaceAudioClassifier',
    'PreTrainedHuggingFaceImageClassifier',
    'PreTrainedHuggingFaceMaskFiller',
    'PreTrainedHuggingFaceObjectDetector',
    'PreTrainedHuggingFaceQuestionAnswerer',
    'PreTrainedHuggingFaceSpeechRecognizer',
    'PreTrainedHuggingFaceTableQuestionAnswerer',
    'PreTrainedHuggingFaceTextClassifier',
    'PreTrainedHuggingFaceTextGenerator',
    'PreTrainedHuggingFaceText2TextGenerator',
    'PreTrainedHuggingFaceTextSummarizer',
    'PreTrainedHuggingFaceTokenClassifier',
    'PreTrainedHuggingFaceTranslator',
    'PreTrainedHuggingFaceZeroShotClassifier',
)



"""Djaizz PIP-related Utilities."""


from sys import version_info
from typing import Optional

from pip._internal.operations.freeze import (   # pylint: disable=import-error
    freeze)

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence   # pylint: disable=ungrouped-imports


__all__: Sequence[str] = ('get_python_dependencies',)


def get_python_dependencies() -> dict[str, Optional[str]]:
    """Get Python Dependencies."""
    deps_and_vers_dict: dict[str, Optional[str]] = {}

    for dep_and_ver_str in freeze(requirement=None,
                                  local_only=False,
                                  user_only=False,
                                  paths=None,
                                  isolated=False,
                                  exclude_editable=False,
                                  skip=()):
        dep_and_ver_tuple: list[str] = dep_and_ver_str.split('==')

        if len(dep_and_ver_tuple) == 2:
            deps_and_vers_dict[dep_and_ver_tuple[0]] = dep_and_ver_tuple[1]

        else:
            assert len(dep_and_ver_tuple) == 1, f'*** {dep_and_ver_tuple} ***'
            deps_and_vers_dict[dep_and_ver_tuple[0]] = None

    return deps_and_vers_dict



"""Djaizz Utilities."""


from importlib import import_module
from sys import version_info
from typing import Any

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence   # pylint: disable=ungrouped-imports


__all__: Sequence[str] = (
    'PGSQL_IDENTIFIER_MAX_LEN',
    'dir_path_with_end_slash',
    'full_qual_name',
    'import_obj',
)


PGSQL_IDENTIFIER_MAX_LEN: int = 63


def dir_path_with_end_slash(path: str) -> str:
    """Add an ending slash to a directory path if necessary."""
    return path if path.endswith('/') else f'{path}/'


def full_qual_name(obj: Any) -> str:
    """full.module.path.and.object.name to fully identify an object."""
    return f'{obj.__module__}.{obj.__qualname__}'


def import_obj(module_and_qualname: str) -> Any:
    """Import an objects by its full.module.path.and.object.name string."""
    module_name, qualname = module_and_qualname.rsplit(sep='.', maxsplit=1)
    return getattr(import_module(module_name), qualname)



"""Djaizz Machine Learning Model base classes."""


from sys import version_info
from typing import Any

from django.db.models.fields import CharField

from djaizz.util import import_obj

from ..base import _AIModelWithArtifactFilesABC

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


__all__: Sequence[str] = ('_PreTrainedMLModelABC',)


class _PreTrainedMLModelABC(_AIModelWithArtifactFilesABC):
    # pylint: disable=abstract-method
    loader_module_and_qualname: CharField = \
        CharField(
            verbose_name='Pre-Trained ML Model Loader (module.dot.qualname)',
            help_text='Pre-Trained ML Model Loader (module.dot.qualname)',

            max_length=255,

            null=False,
            blank=False,
            choices=None,
            db_column=None,
            db_index=True,
            db_tablespace=None,
            default=None,
            editable=True,
            # error_messages={},
            primary_key=False,
            unique=False,
            unique_for_date=None, unique_for_month=None, unique_for_year=None,
            # validators=None
        )

    class Meta(_AIModelWithArtifactFilesABC.Meta):
        # pylint: disable=too-few-public-methods
        """Django Model Class Metadata."""

        abstract = True

    @property
    def loader(self) -> callable:
        """Loader method to load the Model's native object."""
        return import_obj(self.loader_module_and_qualname)

    @property
    def init_params(self) -> dict[str, Any]:
        """Extract initialization parameters from in-database params JSON."""
        return ({} if self.params is None else self.params).get('__init__', {})

    def load(self) -> None:
        if not self.native_obj:
            self.native_obj = self.loader(**self.init_params)



"""Djaizz Pre-Trained Hugging Face Object Detector Model class."""


from sys import version_info
from typing import Any, Union

from django.utils.functional import classproperty

from gradio.interface import Interface
from gradio.inputs import (Image as ImageInput,
                           Slider as SliderInput)
from gradio.outputs import (Image as ImageOutput,   # noqa: N811
                            JSON as JSONOutput)

from PIL.Image import Image

from djaizz.model.apps import DjaizzModelModuleConfig
from djaizz.util import PGSQL_IDENTIFIER_MAX_LEN

from .base import PreTrainedHuggingFaceTransformer

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence   # pylint: disable=ungrouped-imports


__all__: Sequence[str] = ('PreTrainedHuggingFaceObjectDetector',)


ObjectDetectionInputType = Union[str, Image]
ObjectDetectionOutputType = list[dict]


class PreTrainedHuggingFaceObjectDetector(PreTrainedHuggingFaceTransformer):
    # pylint: disable=abstract-method,too-many-ancestors
    """Djaizz Pre-Trained Hugging Face Object Detector Model class."""

    class Meta(PreTrainedHuggingFaceTransformer.Meta):
        # pylint: disable=too-few-public-methods
        """Django Model Class Metadata."""

        verbose_name: str = 'Pre-Trained Hugging Face Object Detector'
        verbose_name_plural: str = 'Pre-Trained Hugging Face Object Detectors'

        db_table: str = (f'{DjaizzModelModuleConfig.label}_'
                         f"{__qualname__.split(sep='.', maxsplit=1)[0]}")
        assert len(db_table) <= PGSQL_IDENTIFIER_MAX_LEN, \
            ValueError(f'*** "{db_table}" DB TABLE NAME TOO LONG ***')

        default_related_name: str = 'pretrained_hugging_face_object_detectors'

    def predict(self,
                image_or_images: Union[ObjectDetectionInputType,
                                       Sequence[ObjectDetectionInputType]],
                threshold: float = 0.9) \
            -> Union[ObjectDetectionOutputType,
                     Sequence[ObjectDetectionOutputType]]:
        # pylint: disable=arguments-differ
        """Detect Objects from Image(s)."""
        if not isinstance(image_or_images, (str, Image, list)):
            image_or_images: list[ObjectDetectionInputType] = \
                list(image_or_images)

        self.load()

        return self.native_obj(inputs=image_or_images,   # images=?
                               threshold=threshold)

    @classproperty
    def gradio_ui(cls) -> Interface:   # noqa: N805
        # pylint: disable=no-self-argument
        """Gradio Interface."""
        BoundingBoxType = tuple[str, int, int, int, int]   # noqa: N806

        def _predict(self, img: Image, threshold: float = 0.9) \
                -> tuple[Image, list[BoundingBoxType]]:
            detected_objs: list[dict[str, Any]] = \
                cls.predict(self, image_or_images=img, threshold=threshold)

            bounding_boxes: list[BoundingBoxType] = []
            for i in detected_objs:
                box_coords = i['box']
                bounding_boxes.append((i['label'],
                                       box_coords['xmin'],   # left-x
                                       box_coords['ymin'],   # top-y
                                       box_coords['xmax'],   # right-x
                                       box_coords['ymax']    # bottom-y
                                       ))

            return (img, bounding_boxes), bounding_boxes

        return Interface(
            fn=_predict,
            # (Callable) - the function to wrap an interface around.

            inputs=[ImageInput(shape=None,
                               image_mode='RGB',
                               invert_colors=False,
                               source='upload',
                               tool='editor',
                               type='pil',
                               label=('Upload an Image to Detect Objects'),
                               optional=False),

                    SliderInput(minimum=.5, maximum=.9, step=.1, default=.9,
                                label='Confidence Threshold')],
            # (Union[str, list[Union[str, InputComponent]]]) -
            # a single Gradio input component,
            # or list of Gradio input components.
            # Components can either be passed as instantiated objects,
            # or referred to by their string shortcuts.
            # The number of input components should match
            # the number of parameters in fn.

            outputs=[ImageOutput(type='pil',
                                 # labeled_segments=True,   # TODO: fix
                                 label='Detected Objects'),

                     JSONOutput(label='Detected Objects')],
            # (Union[str, list[Union[str, OutputComponent]]]) -
            # a single Gradio output component,
            # or list of Gradio output components.
            # Components can either be passed as instantiated objects,
            # or referred to by their string shortcuts.
            # The number of output components should match
            # the number of values returned by fn.

            verbose=True,
            # (bool) - whether to print detailed information during launch.

            examples=None,
            # (Union[list[list[Any]], str]) - sample inputs for the function;
            # if provided, appears below the UI components and can be used
            # to populate the interface.
            # Should be nested list, in which the outer list consists of
            # samples and each inner list consists of an input
            # corresponding to each input component.
            # A string path to a directory of examples can also be provided.
            # If there are multiple input components and a directory
            # is provided, a log.csv file must be present in the directory
            # to link corresponding inputs.

            examples_per_page=10,
            # (int) - If examples are provided, how many to display per page.

            live=False,
            # (bool) - should the interface automatically reload on change?

            layout='unaligned',
            # (str) - Layout of input and output panels.
            # - "horizontal" arranges them as two columns of equal height,
            # - "unaligned" arranges them as two columns of unequal height, and
            # - "vertical" arranges them vertically.

            show_input=True,
            show_output=True,

            capture_session=False,
            # (bool) - if True, captures the default graph and session
            # (needed for Tensorflow 1.x)

            interpretation='default',
            # (Union[Callable, str]) - function that provides interpretation
            # explaining prediction output.
            # Pass "default" to use built-in interpreter.

            num_shap=2.0,
            # (float) - a multiplier that determines how many examples
            # are computed for shap-based interpretation.
            # Increasing this value will increase shap runtime,
            # but improve results.

            theme='default',
            # (str) - Theme to use - one of
            # - "default",
            # - "huggingface",
            # - "grass",
            # - "peach".
            # Add "dark" prefix, e.g. "darkpeach" or "darkdefault"
            # for darktheme.

            repeat_outputs_per_model=True,

            title=cls._meta.verbose_name,
            # (str) - a title for the interface;
            # if provided, appears above the input and output components.

            description='A pre-trained Hugging Face model to detect objects',
            # (str) - a description for the interface;
            # if provided, appears above the input and output components.

            article=None,
            # (str) - an expanded article explaining the interface;
            # if provided, appears below the input and output components.
            # Accepts Markdown and HTML content.

            thumbnail=None,
            # (str) - path to image or src to use as display picture for models
            # listed in gradio.app/hub

            css=None,
            # (str) - custom css or path to custom css file
            # to use with interface.

            server_port=None,
            # (int) - will start gradio app on this port (if available)

            # server_name=networking.LOCALHOST_NAME,
            # (str) - to make app accessible on local network set to "0.0.0.0".

            height=500,
            width=900,

            allow_screenshot=True,
            # (bool) - if False, users will not see a button
            # to take a screenshot of the interface.

            allow_flagging=False,
            # (bool) - if False, users will not see a button
            # to flag an input and output.

            flagging_options=None,
            # (list[str]) - if not None, provides options a user must select
            # when flagging.

            encrypt=False,
            # (bool) - If True, flagged data will be encrypted
            # by key provided by creator at launch

            show_tips=False,
            # (bool) - if True, will occasionally show tips
            # about new Gradio features

            flagging_dir='flagged',
            # (str) - what to name the dir where flagged data is stored.

            analytics_enabled=True,

            enable_queue=False,
            # (bool) - if True, inference requests will be served through
            # a queue instead of with parallel threads.
            # Required for longer inference times (> 1min) to prevent timeout.
        )



"""Djaizz JSON DataSet classes."""


from json.decoder import JSONDecoder   # pylint: disable=import-error
from sys import version_info

from django.core.serializers.json import DjangoJSONEncoder
from django.db.models.fields.json import JSONField

from djaizz.data.apps import DjaizzDataModuleConfig
from djaizz.util import PGSQL_IDENTIFIER_MAX_LEN

from .base import DataSet, _FileDataSetABC

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


__all__: Sequence[str] = ('InDBJSONDataSet', 'JSONDataSet',
                          '_FileDataSetWithInDBJSONCacheABC')


class InDBJSONDataSet(DataSet):
    """Djaizz In-Database JSON DataSet class."""

    in_db_json: JSONField = \
        JSONField(
            verbose_name='In-Database JSON Data Content',
            help_text='In-Database JSON Data Content',

            encoder=DjangoJSONEncoder,
            decoder=JSONDecoder,

            null=True,
            blank=True,
            choices=None,
            db_column=None,
            db_index=False,
            db_tablespace=None,
            default=None,
            editable=True,
            # error_messages={},
            primary_key=False,
            unique=False,
            unique_for_date=None, unique_for_month=None, unique_for_year=None,
            # validators=None
        )

    class Meta(DataSet.Meta):   # pylint: disable=too-few-public-methods
        """Django Model Class Metadata."""

        verbose_name: str = 'In-Database JSON Data Set'
        verbose_name_plural: str = 'In-Database JSON Data Sets'

        db_table: str = (f'{DjaizzDataModuleConfig.label}_'
                         f"{__qualname__.split(sep='.', maxsplit=1)[0]}")
        assert len(db_table) <= PGSQL_IDENTIFIER_MAX_LEN, \
            ValueError(f'*** "{db_table}" DB TABLE NAME TOO LONG ***')

        default_related_name = 'in_db_json_data_sets'


class _FileDataSetWithInDBJSONCacheABC(InDBJSONDataSet, _FileDataSetABC):
    # pylint: disable=abstract-method,too-many-ancestors
    """Djaizz File DataSet with In-Database JSON Cache."""

    class Meta(InDBJSONDataSet.Meta):
        # pylint: disable=too-few-public-methods
        """Django Model Class Metadata."""

        abstract = True


class JSONDataSet(_FileDataSetWithInDBJSONCacheABC):
    # pylint: disable=abstract-method,too-many-ancestors
    """Djaizz JSON DataSet class."""

    class Meta(_FileDataSetWithInDBJSONCacheABC.Meta):
        # pylint: disable=too-few-public-methods
        """Django Model Class Metadata."""

        abstract = True



"""Djaizz AWS Elastic Beanstalk CLI."""


from contextlib import AbstractContextManager
import os
from pathlib import Path
from pprint import pprint
from shutil import copyfile, copytree, ignore_patterns
from typing import Optional

import click
from ruamel import yaml

from ..run_cmd import run_cmd


_DJAI_AWS_EB_CLI_UTIL_DIR_PATH: Path = Path(__file__).parent


_EC2_INSTANCE_TYPES_FILE_NAME: str = 'EC2-Instance-Types.yml'

with open(file=_DJAI_AWS_EB_CLI_UTIL_DIR_PATH / _EC2_INSTANCE_TYPES_FILE_NAME,
          mode='rt',
          buffering=-1,
          encoding='utf-8',
          errors='raise',
          newline=None,
          closefd=True,
          opener=None) as f:
    _EC2_INSTANCE_TYPES: dict = yaml.safe_load(stream=f, version=None)


_EB_EXTENSIONS_DIR_NAME: str = '.ebextensions'
_PLATFORM_DIR_NAME: str = '.platform'
_INSTALL_CUDA_SCRIPT_NAME: str = '.Install-CUDA'


class ConfigFilesHandling(AbstractContextManager):
    """Handle config/extension files, e.g., `.ebextensions`/`.platform`."""

    def __init__(self, config_dir_name: str, /, *, gpu: Optional[bool] = False):  # noqa: E501
        """Initialize context manager."""
        self.config_dir_name: str = config_dir_name
        self.config_dir_path: Path = _DJAI_AWS_EB_CLI_UTIL_DIR_PATH / config_dir_name  # noqa: E501
        self.gpu: bool = gpu

    def __enter__(self):
        """Add config/extension files."""
        copytree(
            src=self.config_dir_path,
            dst=self.config_dir_name,
            symlinks=False,
            ignore=(None
                    if self.gpu
                    else ignore_patterns(_INSTALL_CUDA_SCRIPT_NAME)),
            ignore_dangling_symlinks=False,
            dirs_exist_ok=True)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Remove config/extension files."""
        paths: list[Path] = list(self.config_dir_path.rglob(pattern='*'))

        for path in paths:
            if path.is_file() and (True
                                   if self.gpu
                                   else (path.name != _INSTALL_CUDA_SCRIPT_NAME)):  # noqa: E501
                os.remove(path=path.relative_to(_DJAI_AWS_EB_CLI_UTIL_DIR_PATH))  # noqa: E501

        # remove empty directories
        for path in (paths + [self.config_dir_path]):
            if path.is_dir() and \
                    (not any((dir_path :=
                              path.relative_to(_DJAI_AWS_EB_CLI_UTIL_DIR_PATH))
                             .iterdir())):
                dir_path.rmdir()


class EBIgnoreHandling(AbstractContextManager):
    """Handle `.ebignore` file."""

    EB_IGNORE_FILE_NAME: str = '.ebignore'

    def __init__(self):
        """Initialize context manager."""
        self.eb_ignore_exists: bool = Path(self.EB_IGNORE_FILE_NAME).exists()

        if self.eb_ignore_exists:
            with open(self.EB_IGNORE_FILE_NAME,
                      mode='rt',
                      buffering=-1,
                      encoding='utf-8',
                      errors='strict',
                      newline=None,
                      closefd=True,
                      opener=None) as eb_ignore_file:
                self.eb_ignore_content: list[str] = eb_ignore_file.readlines()

    def __enter__(self):
        """Add `.ebignore` file, if applicable."""
        if self.eb_ignore_exists:
            with (open(self.EB_IGNORE_FILE_NAME,
                       mode='at',
                       buffering=-1,
                       encoding='utf-8',
                       errors='strict',
                       newline=None,
                       closefd=True,
                       opener=None) as dst,
                  open(_DJAI_AWS_EB_CLI_UTIL_DIR_PATH / self.EB_IGNORE_FILE_NAME,  # noqa: E501
                       mode='rt',
                       buffering=-1,
                       encoding='utf-8',
                       errors='strict',
                       newline=None,
                       closefd=True,
                       opener=None) as src):
                dst.writelines(src.readlines())

        else:
            copyfile(
                src=_DJAI_AWS_EB_CLI_UTIL_DIR_PATH / self.EB_IGNORE_FILE_NAME,
                dst=self.EB_IGNORE_FILE_NAME)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Remove `.ebignore` file, if applicable."""
        if self.eb_ignore_exists:
            with open(self.EB_IGNORE_FILE_NAME,
                      mode='wt',
                      buffering=-1,
                      encoding='utf-8',
                      errors='strict',
                      newline=None,
                      closefd=True,
                      opener=None) as eb_ignore_file:
                eb_ignore_file.writelines(self.eb_ignore_content)

        else:
            os.remove(path=self.EB_IGNORE_FILE_NAME)


@click.command(name='init',
               cls=click.Command,
               context_settings=None,
               help=('Djaizz AWS Elastic Beanstalk CLI: '
                     'Initialize Configuration >>>'),
               epilog=('^^^ Djaizz AWS Elastic Beanstalk CLI: '
                       'Initialize Configuration'),
               short_help='Djaizz AWS-EB Init',
               options_metavar='',
               add_help_option=True,
               hidden=False,
               deprecated=False)
def init():
    """Initialize Djaizz AWS Elastic Beanstalk Configuration."""
    os.system(command='eb init')


@click.command(name='deploy',
               cls=click.Command,
               context_settings=None,
               help='Djaizz AWS Elastic Beanstalk CLI: Deploy >>>',
               epilog='^^^ Djaizz AWS Elastic Beanstalk CLI: Deploy',
               short_help='Djaizz AWS-EB Deploy',
               options_metavar='[OPTIONS]',
               add_help_option=True,
               hidden=False,
               deprecated=False)
@click.argument('aws_eb_env_name',
                cls=click.Argument,
                required=False,
                type=str,
                default=None,
                callback=None,
                nargs=None,
                metavar='AWS_EB_ENV_NAME',
                expose_value=True,
                is_eager=False,
                envvar=None)
@click.option('--asgi',
              cls=click.Option,
              show_default=True,
              prompt=False,
              confirmation_prompt=False,
              hide_input=False,
              is_flag=False,
              # flag_value=...,
              multiple=False,
              count=False,
              allow_from_autoenv=False,
              type=str,
              help=('ASGI Server (daphne, hypercorn, uvicorn) to use '
                    '[default: None]'),
              show_choices=True,
              default=None,
              required=False,
              callback=None,
              nargs=None,
              metavar='ASGI',
              expose_value=True,
              is_eager=False,
              envvar=None)
@click.option('--gpu',
              cls=click.Option,
              show_default=True,
              prompt=False,
              confirmation_prompt=False,
              hide_input=False,
              is_flag=True,
              flag_value=True,
              multiple=False,
              count=False,
              allow_from_autoenv=False,
              type=bool,
              help='Whether to use GPU',
              show_choices=True,
              default=False,
              required=False,
              callback=None,
              nargs=None,
              metavar='GPU',
              expose_value=True,
              is_eager=False,
              envvar=None)
def deploy(aws_eb_env_name: Optional[str] = None,
           asgi: Optional[str] = None,
           gpu: Optional[bool] = False):
    """Deploy Djaizz onto AWS Elastic Beanstalk."""
    profile = input('AWS CLI Profile (if not default) = ')
    if not profile.strip():
        profile = 'default'

    with (ConfigFilesHandling(_EB_EXTENSIONS_DIR_NAME),
          ConfigFilesHandling(_PLATFORM_DIR_NAME, gpu=gpu),
          EBIgnoreHandling()):
        if aws_eb_env_name:
            run_cmd(command=f'eb deploy --profile {profile} {aws_eb_env_name}',
                    asgi=asgi)

        else:
            region = input('AWS Region = ')
            vpc = input('AWS VPC = ')
            subnets = input('AWS Subnets = ')
            assert region and vpc and subnets

            # AWS EC2 Instance Type: by default, pick a
            # Compute-optimized instance type
            # with good Networking performance and sufficient Memory
            # (note: Graviton (g) instances not compatible with Djaizz deps)
            if gpu:
                pprint(object=_EC2_INSTANCE_TYPES['gpu'],
                       stream=None,
                       indent=2,
                       width=80,
                       depth=None,
                       compact=True,
                       sort_dicts=False,
                       underscore_numbers=False)
                instance_type = input('AWS EC2 Instance Type '
                                      '(default: g4dn.xlarge; min: g4dn.xlarge) = ')  # noqa: E501
                if not instance_type.strip():
                    instance_type = 'g4dn.xlarge'

            else:
                pprint(object=_EC2_INSTANCE_TYPES['cpu'],
                       stream=None,
                       indent=2,
                       width=80,
                       depth=None,
                       compact=True,
                       sort_dicts=False,
                       underscore_numbers=False)
                instance_type = input('AWS EC2 Instance Type '
                                      '(default: c6i.xlarge; min: c6i.large) = ')  # noqa: E501
                if not instance_type.strip():
                    instance_type = 'c6i.xlarge'

            run_cmd(command=(f'eb create --profile {profile}'
                             f' --region {region}'
                             f' --vpc.id {vpc} --vpc.publicip'
                             f' --vpc.dbsubnets {subnets}'
                             f' --vpc.ec2subnets {subnets}'
                             f' --vpc.elbsubnets {subnets} --vpc.elbpublic'
                             f' --instance_type {instance_type}'
                             f' --timeout 30'),
                    asgi=asgi)


@click.group(name='aws-eb',
             cls=click.Group,
             commands={'init': init, 'deploy': deploy},
             invoke_without_command=False,
             no_args_is_help=True,
             subcommand_metavar='DJAI_AWS_EB_SUB_COMMAND',
             chain=False,
             help='Djaizz AWS Elastic Beanstalk CLI >>>',
             epilog='^^^ Djaizz AWS Elastic Beanstalk CLI',
             short_help='Djaizz AWS-EB CLI',
             options_metavar='[OPTIONS]',
             add_help_option=True,
             hidden=False,
             deprecated=False)
def djaizz_aws_eb():
    """Trigger Djaizz AWS Elastic Beanstalk from CLI."""



"""Djaizz Pre-Trained Hugging Face Model classes."""


from sys import version_info

from .base import PreTrainedHuggingFaceTransformer
from .audio_classification import PreTrainedHuggingFaceAudioClassifier
from .image_classification import PreTrainedHuggingFaceImageClassifier
from .mask_filling import PreTrainedHuggingFaceMaskFiller
from .object_detection import PreTrainedHuggingFaceObjectDetector
from .question_answering import PreTrainedHuggingFaceQuestionAnswerer
from .speech_recognition import PreTrainedHuggingFaceSpeechRecognizer
from .table_question_answering import (
    PreTrainedHuggingFaceTableQuestionAnswerer)
from .text_classification import PreTrainedHuggingFaceTextClassifier
from .text_generation import PreTrainedHuggingFaceTextGenerator
from .text2text_generation import PreTrainedHuggingFaceText2TextGenerator
from .text_summarization import PreTrainedHuggingFaceTextSummarizer
from .token_classification import PreTrainedHuggingFaceTokenClassifier
from .translation import PreTrainedHuggingFaceTranslator
from .zero_shot_classification import PreTrainedHuggingFaceZeroShotClassifier

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


__all__: Sequence[str] = (
    'PreTrainedHuggingFaceTransformer',

    'PreTrainedHuggingFaceAudioClassifier',
    'PreTrainedHuggingFaceImageClassifier',
    'PreTrainedHuggingFaceMaskFiller',
    'PreTrainedHuggingFaceObjectDetector',
    'PreTrainedHuggingFaceQuestionAnswerer',
    'PreTrainedHuggingFaceSpeechRecognizer',
    'PreTrainedHuggingFaceTableQuestionAnswerer',
    'PreTrainedHuggingFaceTextClassifier',
    'PreTrainedHuggingFaceTextGenerator',
    'PreTrainedHuggingFaceText2TextGenerator',
    'PreTrainedHuggingFaceTextSummarizer',
    'PreTrainedHuggingFaceTokenClassifier',
    'PreTrainedHuggingFaceTranslator',
    'PreTrainedHuggingFaceZeroShotClassifier',
)



"""Djaizz Data URLs."""


from sys import version_info

from django.urls.resolvers import URLPattern

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


__all__: Sequence[str] = ('urlpatterns',)


urlpatterns: Sequence[URLPattern] = (
)



"""Djaizz DataSet/DataSource Model classes."""


from sys import version_info

from .base import DataSchema, DataSet

from .json import InDBJSONDataSet, JSONDataSet
from .numpy import NumPyArray
from .pandas import PandasDataFrame

from .audio import AudioDataSet
from .csv import CSVDataSet
from .hdf import HDFDataSet
from .image import ImageDataSet
from .orc import ORCDataSet
from .parquet import ParquetDataSet
from .text import TextDataSet
from .tfrecord import TFRecordDataSet
from .video import VideoDataSet

from .live_api import LiveAPIDataSource

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


__all__: Sequence[str] = (
    'DataSchema', 'DataSet',

    'InDBJSONDataSet', 'JSONDataSet',
    'NumPyArray', 'PandasDataFrame',

    'AudioDataSet',
    'CSVDataSet',
    'HDFDataSet',
    'ImageDataSet',
    'ORCDataSet',
    'ParquetDataSet',
    'TextDataSet',
    'TFRecordDataSet',
    'VideoDataSet',

    'LiveAPIDataSource',
)



"""Set up Google Cloud AI Service models."""


from djaizz.model.models.cloud_ai_svc.google import (
    GoogleTranslate,
)


def run():
    """Run this script."""
    try:
        print(GoogleTranslate.objects
              .get_or_create(name='Google-Translate')[0])

    except Exception as err:   # pylint: disable=broad-except
        print(f'{GoogleTranslate.__name__} model not set up: {err}')



"""Djaizz Cloud AI Service model classes."""


from sys import version_info

from .base import CloudAIService

from .google import (
    GoogleTranslate,
)

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


__all__: Sequence[str] = (
    'CloudAIService',

    'GoogleTranslate',
)



"""Djaizz Dependencies."""


from collections.abc import Sequence
from importlib.metadata import version
from pprint import pprint

import click


CAPPED_DEPS: Sequence[str] = ('Dash',
                              'H11',
                              'NumPy',
                              'Pandas',
                              'Starlette')


@click.command(name='capped-deps',
               cls=click.Command,
               context_settings=None,
               help=('Djaizz Capped Dependencies CLI >>>'),
               epilog=('^^^ Djaizz Capped Dependencies CLI'),
               short_help='Djaizz Capped Dependencies',
               options_metavar='',
               add_help_option=True,
               hidden=False,
               deprecated=False)
def capped_deps():
    """List Djaizz's Capped Dependencies' Versions."""
    pprint(object={dep_name: version(distribution_name=dep_name)
                   for dep_name in CAPPED_DEPS},
           indent=2, width=80, depth=None, compact=False, sort_dicts=False,
           underscore_numbers=True)



"""Djaizz Text DataSet class."""


from sys import version_info

from .base import _FileDataSetABC

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


__all__: Sequence[str] = ('TextDataSet',)


class TextDataSet(_FileDataSetABC):
    # pylint: disable=abstract-method,too-many-ancestors
    """Djaizz Text DataSet class."""

    class Meta(_FileDataSetABC.Meta):
        # pylint: disable=too-few-public-methods
        """Django Model Class Metadata."""

        abstract = True



"""Djaizz Torch Deep Learning Model classes."""


from sys import version_info

from .base import TorchModel

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


__all__: Sequence[str] = (
    'TorchModel',
)



"""Djaizz URLs."""


from sys import version_info

from django.urls.conf import include, path
from django.urls.resolvers import URLPattern

from .data import urls as data_urls
from .model import urls as model_urls

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


__all__: Sequence[str] = ('urlpatterns',)


urlpatterns: Sequence[URLPattern] = (
    path(route='data/',
         view=include(data_urls), kwargs=None, name=None),
    path(route='data-mgmt/',
         view=include(data_urls), kwargs=None, name=None),

    path(route='model/',
         view=include(model_urls), kwargs=None, name=None),
    path(route='models/',
         view=include(model_urls), kwargs=None, name=None),
    path(route='modeling/',
         view=include(model_urls), kwargs=None, name=None),
)



"""Djaizz Pre-Trained Hugging Face Text Generator Model class."""


from sys import version_info
from typing import Optional, Union

from django.utils.functional import classproperty

from gradio.interface import Interface
from gradio.inputs import (Textbox as TextboxInput,
                           Checkbox as CheckboxInput)
from gradio.outputs import JSON as JSONOutput   # noqa: N811

from djaizz.model.apps import DjaizzModelModuleConfig
from djaizz.util import PGSQL_IDENTIFIER_MAX_LEN

from .base import PreTrainedHuggingFaceTransformer

if version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence   # pylint: disable=ungrouped-imports


__all__: Sequence[str] = ('PreTrainedHuggingFaceTextGenerator',)


TextGenerationInputType = str
TextGenerationOutputType = Union[str, list[int]]


class PreTrainedHuggingFaceTextGenerator(PreTrainedHuggingFaceTransformer):
    # pylint: disable=abstract-method,too-many-ancestors
    """Djaizz Pre-Trained Hugging Face Text Generator Model class."""

    class Meta(PreTrainedHuggingFaceTransformer.Meta):
        # pylint: disable=too-few-public-methods
        """Django Model Class Metadata."""

        verbose_name: str = 'Pre-Trained Hugging Face Text Generator'
        verbose_name_plural: str = 'Pre-Trained Hugging Face Text Generators'

        db_table: str = (f'{DjaizzModelModuleConfig.label}_'
                         f"{__qualname__.split(sep='.', maxsplit=1)[0]}")
        assert len(db_table) <= PGSQL_IDENTIFIER_MAX_LEN, \
            ValueError(f'*** "{db_table}" DB TABLE NAME TOO LONG ***')

        default_related_name: str = 'pretrained_hugging_face_text_generators'

    def predict(self,
                text_or_texts: Union[TextGenerationInputType,
                                     Sequence[TextGenerationInputType]],
                return_tensors: bool = False,
                return_text: bool = True,
                return_full_text: bool = True,
                clean_up_tokenization_spaces: bool = True,
                prefix: Optional[str] = None,
                **generate_kwargs) \
            -> Union[TextGenerationOutputType, list[TextGenerationOutputType]]:
        # pylint: disable=arguments-differ,too-many-arguments
        """Generate Text(s)."""
        single_text: bool = isinstance(text_or_texts, str)

        if not (single_text or isinstance(text_or_texts, list)):
            text_or_texts: list[TextGenerationInputType] = list(text_or_texts)

        self.load()

        output = self.native_obj(
            text_inputs=text_or_texts,
            return_tensors=return_tensors,
            return_text=return_text,
            return_full_text=return_full_text,
            clean_up_tokenization_spaces=clean_up_tokenization_spaces,
            prefix=prefix,
            **generate_kwargs)

        if return_tensors:
            # TODO: fix
            # return ((output[0]['generated_token_ids']['output_ids']
            #          .flatten().tolist())
            #         if single_text
            #         else [(result['generated_token_ids']['output_ids']
            #                .flatten().tolist())
            #               for result in output])

            return (output[0]['generated_text']
                    if single_text
                    else [result[0]['generated_text'] for result in output])

        assert return_text, \
            '*** EITHER return_tensors OR return_text MUST BE TRUE ***'

        return (output[0]['generated_text']
                if single_text
                else [result[0]['generated_text'] for result in output])

    @classproperty
    def gradio_ui(cls) -> Interface:   # noqa: N805
        # pylint: disable=no-self-argument
        """Gradio Interface."""
        return Interface(
            fn=cls.predict,
            # (Callable) - the function to wrap an interface around.

            inputs=[TextboxInput(lines=10,
                                 placeholder='Text to Generate from',
                                 default='',
                                 numeric=False,
                                 type='str',
                                 label='Text to Generate from'),

                    CheckboxInput(default=False, label='Return Tensors?'),

                    CheckboxInput(default=True, label='Return Text?'),

                    CheckboxInput(default=True, label='Return Full Text?'),

                    CheckboxInput(default=True, label='Clean Up Spaces?'),

                    TextboxInput(lines=1,
                                 placeholder='Generated Text Prefix',
                                 default='',
                                 numeric=False,
                                 type='str',
                                 label='Generated Text Prefix')],
            # (Union[str, list[Union[str, InputComponent]]]) -
            # a single Gradio input component,
            # or list of Gradio input components.
            # Components can either be passed as instantiated objects,
            # or referred to by their string shortcuts.
            # The number of input components should match
            # the number of parameters in fn.

            outputs=JSONOutput(label='Generated Text'),
            # (Union[str, list[Union[str, OutputComponent]]]) -
            # a single Gradio output component,
            # or list of Gradio output components.
            # Components can either be passed as instantiated objects,
            # or referred to by their string shortcuts.
            # The number of output components should match
            # the number of values returned by fn.

            verbose=True,
            # (bool) - whether to print detailed information during launch.

            examples=None,
            # (Union[list[list[Any]], str]) - sample inputs for the function;
            # if provided, appears below the UI components and can be used
            # to populate the interface.
            # Should be nested list, in which the outer list consists of
            # samples and each inner list consists of an input
            # corresponding to each input component.
            # A string path to a directory of examples can also be provided.
            # If there are multiple input components and a directory
            # is provided, a log.csv file must be present in the directory
            # to link corresponding inputs.

            examples_per_page=10,
            # (int) - If examples are provided, how many to display per page.

            live=False,
            # (bool) - should the interface automatically reload on change?

            layout='unaligned',
            # (str) - Layout of input and output panels.
            # - "horizontal" arranges them as two columns of equal height,
            # - "unaligned" arranges them as two columns of unequal height, and
            # - "vertical" arranges them vertically.

            show_input=True,
            show_output=True,

            capture_session=False,
            # (bool) - if True, captures the default graph and session
            # (needed for Tensorflow 1.x)

            interpretation='default',
            # (Union[Callable, str]) - function that provides interpretation
            # explaining prediction output.
            # Pass "default" to use built-in interpreter.

            num_shap=2.0,
            # (float) - a multiplier that determines how many examples
            # are computed for shap-based interpretation.
            # Increasing this value will increase shap runtime,
            # but improve results.

            theme='default',
            # (str) - Theme to use - one of
            # - "default",
            # - "huggingface",
            # - "grass",
            # - "peach".
            # Add "dark" prefix, e.g. "darkpeach" or "darkdefault"
            # for darktheme.

            repeat_outputs_per_model=True,

            title=cls._meta.verbose_name,
            # (str) - a title for the interface;
            # if provided, appears above the input and output components.

            description=('A pre-trained Hugging Face model '
                         'for Text Generation'),
            # (str) - a description for the interface;
            # if provided, appears above the input and output components.

            article=None,
            # (str) - an expanded article explaining the interface;
            # if provided, appears below the input and output components.
            # Accepts Markdown and HTML content.

            thumbnail=None,
            # (str) - path to image or src to use as display picture for models
            # listed in gradio.app/hub

            css=None,
            # (str) - custom css or path to custom css file
            # to use with interface.

            server_port=None,
            # (int) - will start gradio app on this port (if available)

            # server_name=networking.LOCALHOST_NAME,
            # (str) - to make app accessible on local network set to "0.0.0.0".

            height=500,
            width=900,

            allow_screenshot=True,
            # (bool) - if False, users will not see a button
            # to take a screenshot of the interface.

            allow_flagging=False,
            # (bool) - if False, users will not see a button
            # to flag an input and output.

            flagging_options=None,
            # (list[str]) - if not None, provides options a user must select
            # when flagging.

            encrypt=False,
            # (bool) - If True, flagged data will be encrypted
            # by key provided by creator at launch

            show_tips=False,
            # (bool) - if True, will occasionally show tips
            # about new Gradio features

            flagging_dir='flagged',
            # (str) - what to name the dir where flagged data is stored.

            analytics_enabled=True,

            enable_queue=False,
            # (bool) - if True, inference requests will be served through
            # a queue instead of with parallel threads.
            # Required for longer inference times (> 1min) to prevent timeout.
        )



# pylint: disable=invalid-name


"""Djaizz base DataSchema & DataSet classes."""


import json.decoder
import uuid


import django.core.serializers.json
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone

import model_utils.fields


class Migration(migrations.Migration):
    """Create Djaizz base DataSchema & DataSet classes."""

    initial = True

    dependencies = (('contenttypes', '0002_remove_content_type_name'),)

    operations = (
        migrations.CreateModel(
            name='DataSchema',

            fields=(
                ('polymorphic_ctype',
                 models.ForeignKey(
                    editable=False,
                    null=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    related_name='polymorphic_aidata.dataschema_set+',
                    to='contenttypes.contenttype')),

                ('uuid',
                 models.UUIDField(
                    db_index=True,
                    default=uuid.uuid4,
                    editable=False,
                    help_text='UUID',
                    primary_key=True,
                    serialize=False,
                    unique=True,
                    verbose_name='UUID')),

                ('name',
                 models.CharField(
                    blank=True,
                    db_index=True,
                    default=None,
                    help_text='(optional) Unique Name',
                    max_length=255,
                    null=True,
                    unique=True,
                    verbose_name='(optional) Unique Name')),

                ('specs',
                 models.JSONField(
                    decoder=json.decoder.JSONDecoder,
                    default=None,
                    encoder=django.core.serializers.json.DjangoJSONEncoder,
                    help_text='Data Schema Specifications',
                    verbose_name='Data Schema Specifications')),

                ('created',
                 model_utils.fields.AutoCreatedField(
                    default=django.utils.timezone.now,
                    editable=False,
                    verbose_name='created')),
                ('modified',
                 model_utils.fields.AutoLastModifiedField(
                    default=django.utils.timezone.now,
                    editable=False,
                    verbose_name='modified'))
            ),

            options={
                'verbose_name': 'Data Schema',
                'verbose_name_plural': 'Data Schemas',
                'db_table': 'AIData_DataSchema',
                'ordering': ('name', '-modified'),
                'permissions': (),
                'get_latest_by': 'modified',
                'abstract': False,
                'managed': True,
                'proxy': False,
                'default_permissions': ('add', 'change', 'delete', 'view'),
                'select_on_save': False,
                'default_related_name': 'data_schemas',
                'required_db_features': (),
                'base_manager_name': 'objects',
                'default_manager_name': 'objects'
            }
        ),

        migrations.CreateModel(
            name='DataSet',

            fields=(
                ('polymorphic_ctype',
                 models.ForeignKey(
                    editable=False,
                    null=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    related_name='polymorphic_aidata.dataset_set+',
                    to='contenttypes.contenttype')),

                ('uuid',
                 models.UUIDField(
                    db_index=True,
                    default=uuid.uuid4,
                    editable=False,
                    help_text='UUID',
                    primary_key=True,
                    serialize=False,
                    unique=True,
                    verbose_name='UUID')),

                ('name',
                 models.CharField(
                    blank=True,
                    db_index=True,
                    default=None,
                    help_text='(optional) Unique Name',
                    max_length=255,
                    null=True,
                    unique=True,
                    verbose_name='(optional) Unique Name')),

                ('schema',
                 models.ForeignKey(
                    blank=True,
                    default=None,
                    help_text='Data Set Schema',
                    null=True,
                    on_delete=django.db.models.deletion.SET_NULL,
                    related_name='data_sets',
                    related_query_name='data_set',
                    to='AIData.dataschema',
                    verbose_name='Data Set Schema')),

                ('created',
                 model_utils.fields.AutoCreatedField(
                    default=django.utils.timezone.now,
                    editable=False,
                    verbose_name='created')),
                ('modified',
                 model_utils.fields.AutoLastModifiedField(
                    default=django.utils.timezone.now,
                    editable=False,
                    verbose_name='modified'))
            ),

            options={
                'verbose_name': 'Data Set',
                'verbose_name_plural': 'Data Sets',
                'db_table': 'AIData_DataSet',
                'ordering': ('name', '-modified'),
                'permissions': (),
                'get_latest_by': 'modified',
                'abstract': False,
                'managed': True,
                'proxy': False,
                'default_permissions': ('add', 'change', 'delete', 'view'),
                'select_on_save': False,
                'default_related_name': 'data_sets',
                'required_db_features': (),
                'base_manager_name': 'objects',
                'default_manager_name': 'objects'
            }
        )
    )



# pylint: disable=invalid-name


"""Djaizz PreTrainedHuggingFaceTransformer model classes."""


from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    """Create Djaizz PreTrainedHuggingFaceTransformer model classes."""

    dependencies = (('AIModel', '0002_PreTrainedKerasImageNetClassifier'),)

    operations = (
        migrations.CreateModel(
            name='PreTrainedHuggingFaceTransformer',

            fields=(
                ('aimodel_ptr',
                 models.OneToOneField(
                    auto_created=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    parent_link=True,
                    primary_key=True,
                    related_name='pretrained_hugging_face_transformers',
                    serialize=False, to='AIModel.aimodel')),

                ('artifact_global_url',
                 models.CharField(
                    blank=True,
                    db_index=True,
                    default=None,
                    help_text='Model Artifact Global URL',
                    max_length=255,
                    null=True,
                    unique=True,
                    verbose_name='Model Artifact Global URL')),

                ('artifact_local_path',
                 models.CharField(
                    blank=True,
                    db_index=True,
                    default=None,
                    help_text='Model Artifact Local Path',
                    max_length=255,
                    null=True,
                    verbose_name='Model Artifact Local Path')),

                ('loader_module_and_qualname',
                 models.CharField(
                    db_index=True,
                    default=None,
                    help_text=('Pre-Trained ML Model Loader '
                               '(module.dot.qualname)'),
                    max_length=255,
                    verbose_name=('Pre-Trained ML Model Loader '
                                  '(module.dot.qualname)')))
            ),

            options={
                'verbose_name': 'Pre-Trained Hugging Face Transformer',
                'verbose_name_plural': 'Pre-Trained Hugging Face Transformers',
                'db_table': 'AIModel_PreTrainedHuggingFaceTransformer',
                'abstract': False,
                'default_related_name': 'pretrained_hugging_face_transformers',
                'base_manager_name': 'objects'
            },

            bases=('AIModel.aimodel',)
        ),

        migrations.CreateModel(
            name='PreTrainedHuggingFaceAudioClassifier',

            fields=(
                ('pretrainedhuggingfacetransformer_ptr',
                 models.OneToOneField(
                    auto_created=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    parent_link=True,
                    primary_key=True,
                    related_name='pretrained_hugging_face_audio_classifiers',
                    serialize=False,
                    to='AIModel.pretrainedhuggingfacetransformer')),
            ),

            options={
                'verbose_name': 'Pre-Trained Hugging Face Audio Classifier',
                'verbose_name_plural':
                    'Pre-Trained Hugging Face Audio Classifiers',
                'db_table': 'AIModel_PreTrainedHuggingFaceAudioClassifier',
                'abstract': False,
                'default_related_name':
                    'pretrained_hugging_face_audio_classifiers',
                'base_manager_name': 'objects'
            },

            bases=('AIModel.pretrainedhuggingfacetransformer',)
        ),

        migrations.CreateModel(
            name='PreTrainedHuggingFaceImageClassifier',

            fields=(
                ('pretrainedhuggingfacetransformer_ptr',
                 models.OneToOneField(
                    auto_created=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    parent_link=True,
                    primary_key=True,
                    related_name='pretrained_hugging_face_image_classifiers',
                    serialize=False,
                    to='AIModel.pretrainedhuggingfacetransformer')),
            ),

            options={
                'verbose_name': 'Pre-Trained Hugging Face Image Classifier',
                'verbose_name_plural':
                    'Pre-Trained Hugging Face Image Classifiers',
                'db_table': 'AIModel_PreTrainedHuggingFaceImageClassifier',
                'abstract': False,
                'default_related_name':
                    'pretrained_hugging_face_image_classifiers',
                'base_manager_name': 'objects'
            },

            bases=('AIModel.pretrainedhuggingfacetransformer',)
        ),

        migrations.CreateModel(
            name='PreTrainedHuggingFaceMaskFiller',

            fields=(
                ('pretrainedhuggingfacetransformer_ptr',
                 models.OneToOneField(
                    auto_created=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    parent_link=True,
                    primary_key=True,
                    related_name='pretrained_hugging_face_mask_fillers',
                    serialize=False,
                    to='AIModel.pretrainedhuggingfacetransformer')),
            ),

            options={
                'verbose_name': 'Pre-Trained Hugging Face Mask Filler',
                'verbose_name_plural': 'Pre-Trained Hugging Face Mask Fillers',
                'db_table': 'AIModel_PreTrainedHuggingFaceMaskFiller',
                'abstract': False,
                'default_related_name': 'pretrained_hugging_face_mask_fillers',
                'base_manager_name': 'objects'
            },

            bases=('AIModel.pretrainedhuggingfacetransformer',)
        ),

        migrations.CreateModel(
            name='PreTrainedHuggingFaceObjectDetector',

            fields=(
                ('pretrainedhuggingfacetransformer_ptr',
                 models.OneToOneField(
                    auto_created=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    parent_link=True,
                    primary_key=True,
                    related_name='pretrained_hugging_face_object_detectors',
                    serialize=False,
                    to='AIModel.pretrainedhuggingfacetransformer')),
            ),

            options={
                'verbose_name': 'Pre-Trained Hugging Face Object Detector',
                'verbose_name_plural':
                    'Pre-Trained Hugging Face Object Detectors',
                'db_table': 'AIModel_PreTrainedHuggingFaceObjectDetector',
                'abstract': False,
                'default_related_name':
                    'pretrained_hugging_face_object_detectors',
                'base_manager_name': 'objects'
            },

            bases=('AIModel.pretrainedhuggingfacetransformer',)
        ),

        migrations.CreateModel(
            name='PreTrainedHuggingFaceQuestionAnswerer',

            fields=(
                ('pretrainedhuggingfacetransformer_ptr',
                 models.OneToOneField(
                    auto_created=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    parent_link=True,
                    primary_key=True,
                    related_name='pretrained_hugging_face_question_answerers',
                    serialize=False,
                    to='AIModel.pretrainedhuggingfacetransformer')),
            ),

            options={
                'verbose_name': 'Pre-Trained Hugging Face Question Answerer',
                'verbose_name_plural':
                    'Pre-Trained Hugging Face Question Answerers',
                'db_table': 'AIModel_PreTrainedHuggingFaceQuestionAnswerer',
                'abstract': False,
                'default_related_name':
                    'pretrained_hugging_face_question_answerers',
                'base_manager_name': 'objects'
            },

            bases=('AIModel.pretrainedhuggingfacetransformer',),
        ),
        migrations.CreateModel(
            name='PreTrainedHuggingFaceSpeechRecognizer',

            fields=(
                ('pretrainedhuggingfacetransformer_ptr',
                 models.OneToOneField(
                    auto_created=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    parent_link=True,
                    primary_key=True,
                    related_name='pretrained_hugging_face_speech_recognizers',
                    serialize=False,
                    to='AIModel.pretrainedhuggingfacetransformer')),
            ),

            options={
                'verbose_name': 'Pre-Trained Hugging Face Speech Recognizer',
                'verbose_name_plural':
                    'Pre-Trained Hugging Face Speech Recognizers',
                'db_table': 'AIModel_PreTrainedHuggingFaceSpeechRecognizer',
                'abstract': False,
                'default_related_name':
                    'pretrained_hugging_face_speech_recognizers',
                'base_manager_name': 'objects'
            },

            bases=('AIModel.pretrainedhuggingfacetransformer',)
        ),

        migrations.CreateModel(
            name='PreTrainedHuggingFaceTableQuestionAnswerer',

            fields=(
                ('pretrainedhuggingfacetransformer_ptr',
                 models.OneToOneField(
                    auto_created=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    parent_link=True,
                    primary_key=True,
                    related_name=('pretrained_hugging_face_'
                                  'table_question_answerers'),
                    serialize=False,
                    to='AIModel.pretrainedhuggingfacetransformer')),
            ),

            options={
                'verbose_name':
                    'Pre-Trained Hugging Face Table Question Answerer',
                'verbose_name_plural':
                    'Pre-Trained Hugging Face Table Question Answerers',
                'db_table':
                    'AIModel_PreTrainedHuggingFaceTableQuestionAnswerer',
                'abstract': False,
                'default_related_name':
                    'pretrained_hugging_face_table_question_answerers',
                'base_manager_name': 'objects'
            },

            bases=('AIModel.pretrainedhuggingfacetransformer',)
        ),

        migrations.CreateModel(
            name='PreTrainedHuggingFaceText2TextGenerator',

            fields=(
                ('pretrainedhuggingfacetransformer_ptr',
                 models.OneToOneField(
                    auto_created=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    parent_link=True,
                    primary_key=True,
                    related_name=('pretrained_hugging_face_'
                                  'text2text_generators'),
                    serialize=False,
                    to='AIModel.pretrainedhuggingfacetransformer')),
            ),

            options={
                'verbose_name':
                    'Pre-Trained Hugging Face Text-to-Text Generator',
                'verbose_name_plural':
                    'Pre-Trained Hugging Face Text-to-Text Generators',
                'db_table': 'AIModel_PreTrainedHuggingFaceText2TextGenerator',
                'abstract': False,
                'default_related_name':
                    'pretrained_hugging_face_text2text_generators',
                'base_manager_name': 'objects'
            },

            bases=('AIModel.pretrainedhuggingfacetransformer',)
        ),

        migrations.CreateModel(
            name='PreTrainedHuggingFaceTextClassifier',

            fields=(
                ('pretrainedhuggingfacetransformer_ptr',
                 models.OneToOneField(
                    auto_created=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    parent_link=True,
                    primary_key=True,
                    related_name='pretrained_hugging_face_text_classifiers',
                    serialize=False,
                    to='AIModel.pretrainedhuggingfacetransformer')),
            ),

            options={
                'verbose_name': 'Pre-Trained Hugging Face Text Classifier',
                'verbose_name_plural':
                    'Pre-Trained Hugging Face Text Classifiers',
                'db_table': 'AIModel_PreTrainedHuggingFaceTextClassifier',
                'abstract': False,
                'default_related_name':
                    'pretrained_hugging_face_text_classifiers',
                'base_manager_name': 'objects'
            },

            bases=('AIModel.pretrainedhuggingfacetransformer',)
        ),

        migrations.CreateModel(
            name='PreTrainedHuggingFaceTextGenerator',

            fields=(
                ('pretrainedhuggingfacetransformer_ptr',
                 models.OneToOneField(
                    auto_created=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    parent_link=True,
                    primary_key=True,
                    related_name='pretrained_hugging_face_text_generators',
                    serialize=False,
                    to='AIModel.pretrainedhuggingfacetransformer')),
            ),

            options={
                'verbose_name': 'Pre-Trained Hugging Face Text Generator',
                'verbose_name_plural':
                    'Pre-Trained Hugging Face Text Generators',
                'db_table': 'AIModel_PreTrainedHuggingFaceTextGenerator',
                'abstract': False,
                'default_related_name':
                    'pretrained_hugging_face_text_generators',
                'base_manager_name': 'objects'
            },

            bases=('AIModel.pretrainedhuggingfacetransformer',)
        ),

        migrations.CreateModel(
            name='PreTrainedHuggingFaceTextSummarizer',

            fields=(
                ('pretrainedhuggingfacetransformer_ptr',
                 models.OneToOneField(
                    auto_created=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    parent_link=True,
                    primary_key=True,
                    related_name='pretrained_hugging_face_text_summarizers',
                    serialize=False,
                    to='AIModel.pretrainedhuggingfacetransformer')),
            ),

            options={
                'verbose_name': 'Pre-Trained Hugging Face Text Summarizer',
                'verbose_name_plural':
                    'Pre-Trained Hugging Face Text Summarizers',
                'db_table': 'AIModel_PreTrainedHuggingFaceTextSummarizer',
                'abstract': False,
                'default_related_name':
                    'pretrained_hugging_face_text_summarizers',
                'base_manager_name': 'objects'
            },

            bases=('AIModel.pretrainedhuggingfacetransformer',)
        ),

        migrations.CreateModel(
            name='PreTrainedHuggingFaceTokenClassifier',

            fields=(
                ('pretrainedhuggingfacetransformer_ptr',
                 models.OneToOneField(
                    auto_created=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    parent_link=True,
                    primary_key=True,
                    related_name='pretrained_hugging_face_token_classifiers',
                    serialize=False,
                    to='AIModel.pretrainedhuggingfacetransformer')),
            ),

            options={
                'verbose_name': 'Pre-Trained Hugging Face Token Classifier',
                'verbose_name_plural':
                    'Pre-Trained Hugging Face Token Classifiers',
                'db_table': 'AIModel_PreTrainedHuggingFaceTokenClassifier',
                'abstract': False,
                'default_related_name':
                    'pretrained_hugging_face_token_classifiers',
                'base_manager_name': 'objects'
            },

            bases=('AIModel.pretrainedhuggingfacetransformer',)
        ),

        migrations.CreateModel(
            name='PreTrainedHuggingFaceTranslator',

            fields=(
                ('pretrainedhuggingfacetransformer_ptr',
                 models.OneToOneField(
                    auto_created=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    parent_link=True,
                    primary_key=True,
                    related_name='pretrained_hugging_face_translators',
                    serialize=False,
                    to='AIModel.pretrainedhuggingfacetransformer')),
            ),

            options={
                'verbose_name': 'Pre-Trained Hugging Face Translator',
                'verbose_name_plural': 'Pre-Trained Hugging Face Translators',
                'db_table': 'AIModel_PreTrainedHuggingFaceTranslator',
                'abstract': False,
                'default_related_name': 'pretrained_hugging_face_translators',
                'base_manager_name': 'objects'
            },

            bases=('AIModel.pretrainedhuggingfacetransformer',)
        ),

        migrations.CreateModel(
            name='PreTrainedHuggingFaceZeroShotClassifier',

            fields=(
                ('pretrainedhuggingfacetransformer_ptr',
                 models.OneToOneField(
                    auto_created=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    parent_link=True,
                    primary_key=True,
                    related_name=('pretrained_hugging_face_'
                                  'zero_shot_classifiers'),
                    serialize=False,
                    to='AIModel.pretrainedhuggingfacetransformer')),
            ),

            options={
                'verbose_name':
                    'Pre-Trained Hugging Face Zero-Shot Classifier',
                'verbose_name_plural':
                    'Pre-Trained Hugging Face Zero-Shot Classifiers',
                'db_table': 'AIModel_PreTrainedHuggingFaceZeroShotClassifier',
                'abstract': False,
                'default_related_name':
                    'pretrained_hugging_face_zero_shot_classifiers',
                'base_manager_name': 'objects'
            },

            bases=('AIModel.pretrainedhuggingfacetransformer',)
        )
    )
